# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>

from collections.abc import Iterable
from datetime import datetime

import pytest

from odfdo.document import Document
from odfdo.element import Element
from odfdo.note import Annotation, AnnotationEnd
from odfdo.paragraph import Paragraph


@pytest.fixture
def document(samples) -> Iterable[Document]:
    document = Document(samples("note.odt"))
    yield document


def test_get_office_names_0():
    doc = Document()
    assert doc.body.get_office_names() == []


def test_get_office_names_1(document):
    assert document.body.get_office_names() == []


def test_get_office_names_2(document):
    annotation = Annotation(
        "Some annotation",
        creator="John Doe",
        date=datetime(2025, 6, 7, 12, 00, 00),
    )
    document.body.append(annotation)
    assert document.body.get_office_names() == ["__Fieldmark__lpod_1"]


def test_get_office_names_3(document):
    annotation1 = Annotation(
        "Some annotation 1",
        creator="John Doe",
        date=datetime(2025, 6, 7, 12, 00, 00),
        parent=document.body,
    )
    document.body.append(annotation1)
    annotation2 = Annotation(
        "Some annotation 1",
        creator="John Doe",
        date=datetime(2025, 6, 7, 12, 00, 00),
        parent=document.body,
    )
    document.body.append(annotation2)
    assert set(document.body.get_office_names()) == {
        "__Fieldmark__lpod_1",
        "__Fieldmark__lpod_2",
    }


def test_annotation_class():
    annotation = Annotation()
    assert isinstance(annotation, Annotation)


def test_annotation_minimal_tag():
    annotation = Element.from_tag("<office:annotation/>")
    assert isinstance(annotation, Annotation)


def test_annotation_with_name():
    annotation = Annotation(name="some name")
    assert annotation.name == "some name"


def test_annotation_set_dc_creator():
    annotation = Annotation()
    name = "John Doe"
    annotation.dc_creator = name
    assert annotation.dc_creator == name


def test_annotation_set_dc_date():
    annotation = Annotation()
    dt = datetime(2024, 12, 25, 12, 0, 0)
    annotation.dc_date = dt
    assert annotation.dc_date == dt


def test_annotation_set_note_body_str():
    annotation = Annotation()
    content = "Some string content"
    annotation.note_body = content
    assert annotation.note_body == content


def test_annotation_set_note_body_element():
    annotation = Annotation()
    content = Paragraph("Some element content")
    annotation.note_body = content
    assert annotation.note_body == str(content).strip()


def test_annotation_set_note_body_wrong_type():
    annotation = Annotation()
    with pytest.raises(TypeError):
        annotation.note_body = []


def test_annotation_start():
    annotation = Annotation()
    assert annotation.start == annotation


def test_annotation_end():
    annotation = Annotation()
    assert annotation.end is None


def test_annotation_end_minimal():
    annotation = Element.from_tag("<office:annotation/>")
    assert annotation.end is None


def test_annotation_end_when_body(document):
    body = document.body
    annotation = body.get_annotations()[0]
    assert annotation.end is None


def test_annotation_get_annotated_none(document):
    body = document.body
    annotation = body.get_annotations()[0]
    result = annotation.get_annotated()
    assert result is None


def test_annotation_get_annotated_str(document):
    body = document.body
    annotation = body.get_annotations()[0]
    result = annotation.get_annotated(as_text=True)
    assert result == ""


def test_insert_annotation_with_end():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    paragraph.insert_annotation_end(annotation, after="graphe")
    expected = (
        "<text:p>Un para"
        '<office:annotation office:name="__Fieldmark__lpod_1">'
        "<text:p>It's like you're in a cave.</text:p>"
        "<dc:creator>Plato</dc:creator><dc:date>2025-06-07T00:00:00</dc:date>"
        "</office:annotation>"
        "graphe"
        '<office:annotation-end office:name="__Fieldmark__lpod_1"/>'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_insertget_long_annotated():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    paragraph.insert_annotation_end(annotation, after="graphe")
    result = annotation.get_annotated()
    assert len(result) == 1
    assert isinstance(result[0], Paragraph)


def test_insertget_long_annotated_str():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    paragraph.insert_annotation_end(annotation, after="graphe")
    result = annotation.get_annotated(as_text=True)
    assert result == "graphe\n\n"


def test_insertget_short_delete():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    annotation.delete()
    annotations = paragraph.get_annotations()
    assert annotations == []


def test_insertget_long_delete():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    paragraph.insert_annotation_end(annotation, after="graphe")
    annotation.delete()
    annotations = paragraph.get_annotations()
    assert annotations == []
    annotations_end = paragraph.get_annotation_ends()
    assert annotations_end == []


def test_annotations_no_date_by_delete():
    text = "It's like you're in a cave."
    creator = "Plato"
    annotation = Annotation(text, creator=creator)
    # date automatically created, remove:
    adate = annotation.get_element("//dc:date")
    adate.delete()
    assert annotation.serialize() == (
        "<office:annotation "
        'office:name="__Fieldmark__lpod_1">'
        "<text:p>It's like you're in a cave.</text:p>"
        "<dc:creator>Plato</dc:creator>"
        "</office:annotation>"
    )


def test_annotations_no_date_get_annotations():
    text = "It's like you're in a cave."
    creator = "Plato"
    annotation = Annotation(text, creator=creator)
    # date automatically created, remove:
    adate = annotation.get_element("//dc:date")
    adate.delete()
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    annotations = paragraph.get_annotations()
    # date is created again
    assert len(annotations) == 1


def test_get_annotation_1():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    annotation = paragraph.get_annotation()
    assert annotation.creator == "Plato"


def test_get_annotation_2_none_position():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    annotation = paragraph.get_annotation(5)
    assert annotation is None


def test_get_annotation_3_creator():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    annotation = paragraph.get_annotation(creator="Plato")
    assert annotation.creator == "Plato"


def test_get_annotation_4_none_creator():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    annotation = paragraph.get_annotation(creator="Socrates")
    assert annotation is None


def test_annotation_check_valid_no_note_body_0():
    annotation = Element.from_tag("<office:annotation/>")
    with pytest.raises(ValueError):
        annotation.check_validity()


def test_annotation_check_valid_no_note_body_1():
    annotation = Annotation()
    with pytest.raises(ValueError):
        annotation.check_validity()


def test_annotation_check_valid_no_dc_creator():
    annotation = Annotation("Some body")
    with pytest.raises(ValueError):
        annotation.check_validity()


def test_annotation_check_valid_no_dc_date():
    annotation = Annotation("Some body", creator="John")
    annotation.dc_date = None
    assert annotation.check_validity() is None


def test_annotation_check_valid_no_dc_date_2():
    annotation = Annotation("Some body", creator="John")
    to_del = annotation.get_element("//dc:date")
    annotation.delete(to_del)
    assert annotation.check_validity() is None


def test_create_annotation():
    # Create
    annotation = Annotation(
        "Lost Dialogs",
        creator="Plato",
        date=datetime(2009, 6, 22, 17, 18, 42),
    )
    expected = (
        '<office:annotation office:name="__Fieldmark__lpod_1">'
        "<text:p>"
        "Lost Dialogs"
        "</text:p>"
        "<dc:creator>Plato</dc:creator>"
        "<dc:date>2009-06-22T17:18:42</dc:date>"
        "</office:annotation>"
    )
    assert annotation.serialize() == expected


def test_create_annotation_str():
    # Create
    annotation = Annotation(
        "Lost Dialogs",
        creator="Plato",
        date=datetime(2009, 6, 22, 17, 18, 42),
    )
    expected = "Lost Dialogs\nPlato 2009-06-22 17:18:42"
    assert str(annotation) == expected


def test_get_annotation_list(document):
    body = document.body
    annotations = body.get_annotations()
    assert len(annotations) == 1
    annotation = annotations[0]
    creator = annotation.dc_creator
    assert creator == "Auteur inconnu"
    date = annotation.dc_date
    assert date == datetime(2009, 8, 3, 12, 9, 45)
    text = annotation.text_content
    expected = "Sauf qu'il est commenté !"
    assert text == expected


def test_get_annotation_list_properties_creator(document):
    body = document.body
    annotations = body.get_annotations()
    annotation = annotations[0]
    creator = annotation.creator
    assert creator == "Auteur inconnu"


def test_get_annotation_list_properties_date(document):
    body = document.body
    annotations = body.get_annotations()
    annotation = annotations[0]
    date = annotation.date
    assert date == datetime(2009, 8, 3, 12, 9, 45)


def test_get_annotation_list_author(document):
    body = document.body
    creator = "Auteur inconnu"
    annotations = body.get_annotations(creator=creator)
    assert len(annotations) == 1


def test_get_annotation_list_bad_author(document):
    body = document.body
    creator = "Plato"
    annotations = body.get_annotations(creator=creator)
    assert len(annotations) == 0


def test_get_annotation_list_start_date(document):
    body = document.body
    start_date = datetime(2009, 8, 1, 0, 0, 0)
    annotations = body.get_annotations(start_date=start_date)
    assert len(annotations) == 1


def test_get_annotation_list_bad_start_date(document):
    body = document.body
    start_date = datetime(2009, 9, 1, 0, 0, 0)
    annotations = body.get_annotations(start_date=start_date)
    assert len(annotations) == 0


def test_get_annotation_list_end_date(document):
    body = document.body
    end_date = datetime(2009, 9, 1, 0, 0, 0)
    annotations = body.get_annotations(end_date=end_date)
    assert len(annotations) == 1


def test_get_annotation_list_bad_end_date(document):
    body = document.body
    end_date = datetime(2009, 8, 1, 0, 0, 0)
    annotations = body.get_annotations(end_date=end_date)
    assert len(annotations) == 0


def test_get_annotation_list_start_date_end_date(document):
    body = document.body
    start_date = datetime(2009, 8, 1, 0, 0, 0)
    end_date = datetime(2009, 9, 1, 0, 0, 0)
    annotations = body.get_annotations(start_date=start_date, end_date=end_date)
    assert len(annotations) == 1


def test_get_annotation_list_bad_start_date_end_date(document):
    body = document.body
    start_date = datetime(2009, 5, 1, 0, 0, 0)
    end_date = datetime(2009, 6, 1, 0, 0, 0)
    annotations = body.get_annotations(start_date=start_date, end_date=end_date)
    assert len(annotations) == 0


def test_get_annotation_list_author_start_date_end_date(document):
    body = document.body
    creator = "Auteur inconnu"
    start_date = datetime(2009, 8, 1, 0, 0, 0)
    end_date = datetime(2009, 9, 1, 0, 0, 0)
    annotations = body.get_annotations(
        creator=creator, start_date=start_date, end_date=end_date
    )
    assert len(annotations) == 1


def test_get_annotation_list_bad_author_start_date_end_date(document):
    body = document.body
    creator = "Plato"
    start_date = datetime(2009, 6, 1, 0, 0, 0)
    end_date = datetime(2009, 7, 1, 0, 0, 0)
    annotations = body.get_annotations(
        creator=creator, start_date=start_date, end_date=end_date
    )
    assert len(annotations) == 0


def test_get_annotation_list_author_bad_start_date_end_date(document):
    body = document.body
    creator = "Auteur inconnu"
    start_date = datetime(2009, 6, 23, 0, 0, 0)
    end_date = datetime(2009, 7, 1, 0, 0, 0)
    annotations = body.get_annotations(
        creator=creator, start_date=start_date, end_date=end_date
    )
    assert len(annotations) == 0


def test_insert_annotation():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2009, 8, 19)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    expected = (
        "<text:p>Un para"
        '<office:annotation office:name="__Fieldmark__lpod_1">'
        "<text:p>It's like you're in a cave.</text:p>"
        "<dc:creator>Plato</dc:creator>"
        "<dc:date>2009-08-19T00:00:00</dc:date>"
        "</office:annotation>"
        "graphe</text:p>"
    )
    assert paragraph.serialize() == expected


def test_annotation_end_class():
    with pytest.raises(ValueError):
        AnnotationEnd()


def test_annotation_end_class_1():
    annotation_end = AnnotationEnd(name="some name")
    assert isinstance(annotation_end, AnnotationEnd)


def test_annotation_end_minimal_tag():
    annotation_end = Element.from_tag("<office:annotation-end/>")
    assert isinstance(annotation_end, AnnotationEnd)


def test_annotation_end_from_annotation_start():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    annotation_end = paragraph.insert_annotation_end(annotation, after="graphe")
    assert annotation_end.start.name == annotation.name


def test_annotation_end_from_annotation_end():
    text = "It's like you're in a cave."
    creator = "Plato"
    date = datetime(2025, 6, 7)
    annotation = Annotation(text, creator=creator, date=date)
    paragraph = Paragraph("Un paragraphe")
    paragraph.insert_annotation(annotation, after="para")
    annotation_end = paragraph.insert_annotation_end(annotation, after="graphe")
    assert annotation_end.end == annotation_end



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from collections.abc import Iterable
from decimal import Decimal

import pytest

from odfdo.const import ODF_CONTENT
from odfdo.container import Container
from odfdo.element import Element
from odfdo.table import Table
from odfdo.xmlpart import XmlPart


@pytest.fixture
def chart_container(samples) -> Iterable[Container]:
    container = Container()
    container.open(samples("chart.odt"))
    yield container


def test_get_element_list(chart_container):
    content_part = XmlPart(ODF_CONTENT, chart_container)
    elements = content_part.get_elements("//text:p")
    assert len(elements) == 4


def test_get_body(chart_container):
    content_part = XmlPart(ODF_CONTENT, chart_container)
    body = content_part.body
    assert body.tag == "office:text"


def test_parts(chart_container):
    parts = chart_container.get_parts()
    assert "mimetype" in parts
    assert "content.xml" in parts
    assert "Object 1/content.xml" in parts


def test_parts_property(chart_container):
    parts = chart_container.parts
    assert "mimetype" in parts
    assert "content.xml" in parts
    assert "Object 1/content.xml" in parts


def test_content_chart(chart_container):
    obj = XmlPart("Object 1/content.xml", chart_container)
    body = obj.body
    assert body is not None
    assert isinstance(body, Element)


def test_rows_columns(chart_container):
    obj = XmlPart("Object 1/content.xml", chart_container)
    body = obj.body
    table = body.get_table(0)
    assert isinstance(table, Table)
    print(table.serialize(pretty=True))
    assert table.width == 4
    assert table.height == 5


def test_read_values(chart_container):
    obj = XmlPart("Object 1/content.xml", chart_container)
    body = obj.body
    table = body.get_table(0)
    values = table.get_values()
    # remove "NaN" results:
    values[1][1] = None
    values[2][1] = None
    values[3][1] = None
    values[4][1] = None
    assert values[0] == [None, "", "Column 2", "Column 3"]
    assert values[1] == ["Row 1", None, 10, 20]
    assert values[2] == ["Row 2", None, 30, 40]
    assert values[3] == ["Row 3", None, 50, 360]
    assert values[4] == ["Row 4", None, Decimal("9.02"), Decimal("6.2")]


def test_change_value(chart_container):
    obj = XmlPart("Object 1/content.xml", chart_container)
    body = obj.body
    table = body.get_table(0)
    table.set_value("A2", "row1 changed")
    table.set_value("D3", 4000)
    values = table.get_values()
    # remove "NaN" results:
    values[1][1] = None
    values[2][1] = None
    values[3][1] = None
    values[4][1] = None
    assert values[0] == [None, "", "Column 2", "Column 3"]
    assert values[1] == ["row1 changed", None, 10, 20]
    assert values[2] == ["Row 2", None, 30, 4000]
    assert values[3] == ["Row 3", None, 50, 360]
    assert values[4] == ["Row 4", None, Decimal("9.02"), Decimal("6.2")]



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>

import io
from collections.abc import Iterable
from datetime import date as dtdate
from datetime import datetime, timedelta
from decimal import Decimal
from textwrap import dedent

import pytest

from odfdo import Meta
from odfdo.body import Metadata
from odfdo.const import ODF_META
from odfdo.datatype import DateTime, Duration
from odfdo.document import Document
from odfdo.meta import GENERATOR
from odfdo.meta_auto_reload import MetaAutoReload
from odfdo.meta_hyperlink_behaviour import MetaHyperlinkBehaviour
from odfdo.meta_template import MetaTemplate


@pytest.fixture
def meta(samples) -> Iterable[Meta]:
    document = Document(samples("meta.odt"))
    yield document.get_part(ODF_META)


def test_get_title(meta):
    title = meta.get_title()
    expected = "Intitulé"
    assert title == expected


def test_set_title(meta):
    clone = meta.clone
    title = "Nouvel intitulé"
    clone.set_title(title)
    assert clone.get_title() == title


def test_title_property(meta):
    clone = meta.clone
    title = "Nouvel intitulé"
    clone.title = title
    assert clone.title == title


def test_title_property_empty(meta):
    clone = meta.clone
    orig = clone.get_element("//dc:title")
    orig.delete()
    title = "Nouvel intitulé"
    clone.title = title
    assert clone.title == title


def test_get_description(meta):
    description = meta.get_description()
    expected = "Comments\nCommentaires\n评论"
    assert description == expected


def test_set_description(meta):
    clone = meta.clone
    description = "评论\nnCommentaires\nComments"
    clone.set_description(description)
    assert clone.get_description() == description


def test_description_property(meta):
    clone = meta.clone
    description = "评论\nnCommentaires\nComments"
    clone.description = description
    assert clone.description == description


def test_description_property_empty(meta):
    clone = meta.clone
    orig = clone.get_element("//dc:description")
    orig.delete()
    description = "评论\nnCommentaires\nComments"
    clone.description = description
    assert clone.description == description


def test_get_subject(meta):
    subject = meta.get_subject()
    expected = "Sujet de sa majesté"
    assert subject == expected


def test_set_subject(meta):
    clone = meta.clone
    subject = "Θέμα"
    clone.set_subject(subject)
    assert clone.get_subject() == subject


def test_subject_property(meta):
    clone = meta.clone
    subject = "Θέμα"
    clone.subject = subject
    assert clone.subject == subject


def test_subject_property_empty(meta):
    clone = meta.clone
    orig = clone.get_element("//dc:subject")
    orig.delete()
    subject = "Θέμα"
    clone.subject = subject
    assert clone.subject == subject


def test_get_language(meta):
    language = meta.get_language()
    assert language is None


def test_set_language(meta):
    clone = meta.clone
    language = "en-US"
    clone.set_language(language)
    assert clone.get_language() == language


def test_set_bad_language(meta):
    clone = meta.clone
    language = "English"
    with pytest.raises(TypeError):
        clone.set_language(language)


def test_language_property(meta):
    clone = meta.clone
    language = "en-US"
    clone.language = language
    assert clone.language == language
    assert clone.get_language() == language


def test_language_property_double(meta):
    clone = meta.clone
    language = "en-US"
    language2 = "en-UK"
    clone.language = language
    assert clone.language == language
    clone.language = language2
    assert clone.language == language2


def test_get_modification_date(meta):
    date = meta.get_modification_date()
    expected = DateTime.decode("2009-07-31T15:59:13")
    assert date == expected
    assert meta.date == expected


def test_set_modification_date(meta):
    clone = meta.clone
    now = datetime.now().replace(microsecond=0)
    clone.set_modification_date(now)
    assert clone.get_modification_date() == now


def test_set_modification_date_meta_special(meta):
    clone = meta.clone
    elem = clone.get_element("//dc:date")
    elem.delete()
    now = datetime.now().replace(microsecond=0)
    clone.set_modification_date(now)
    assert clone.get_modification_date() == now


def test_modification_date_property(meta):
    clone = meta.clone
    now = datetime.now().replace(microsecond=0)
    clone.date = now
    assert clone.date == now
    assert clone.get_modification_date() == now


def test_set_bad_modication_date(meta):
    clone = meta.clone
    date = "2009-06-29T14:15:45"
    with pytest.raises(AttributeError):
        clone.set_modification_date(date)


def test_print_date_property(meta):
    clone = meta.clone
    now = datetime.now().replace(microsecond=0)
    clone.print_date = now
    assert clone.print_date == now


def test_print_date_property_2(meta):
    clone = meta.clone
    now = datetime.now().replace(microsecond=0)
    clone.print_date = now
    # second change
    now = now + timedelta(hours=1)
    clone.print_date = now
    assert clone.print_date == now


def test_print_date_property_none(meta):
    clone = meta.clone
    now = datetime.now().replace(microsecond=0)
    clone.print_date = None
    assert clone.print_date - now < timedelta(hours=3)


def test_get_creation_date(meta):
    date = meta.get_creation_date()
    expected = datetime(2009, 7, 31, 15, 57, 37)
    assert date == expected


def test_set_creation_date(meta):
    clone = meta.clone
    now = datetime.now().replace(microsecond=0)
    clone.set_creation_date(now)
    assert clone.get_creation_date() == now


def test_set_bad_creation_date(meta):
    clone = meta.clone
    date = "2009-06-29T14:15:45"
    with pytest.raises(AttributeError):
        clone.set_creation_date(date)


def test_set_creation_date_none(meta):
    clone = meta.clone
    now_date = datetime.now()
    clone.set_creation_date(None)
    assert clone.get_creation_date() - now_date < timedelta(hours=3)


def test_creation_date_property(meta):
    clone = meta.clone
    now = datetime.now().replace(microsecond=0)
    clone.creation_date = now
    assert clone.creation_date == now


def test_get_initial_creator(meta):
    creator = meta.get_initial_creator()
    expected = None
    assert creator == expected


def test_set_initial_creator(meta):
    clone = meta.clone
    creator = "Hervé"
    clone.set_initial_creator(creator)
    assert clone.get_initial_creator() == creator


def test_initial_creator_property(meta):
    clone = meta.clone
    creator = "Hervé"
    clone.initial_creator = creator
    assert clone.initial_creator == creator


def test_initial_creator_property_twice(meta):
    clone = meta.clone
    creator = "Hervé"
    creator2 = "John"
    clone.initial_creator = creator
    clone.initial_creator = creator2
    assert clone.initial_creator == creator2


def test_printed_by_property(meta):
    clone = meta.clone
    printer = "Jérôme"
    clone.printed_by = printer
    assert clone.printed_by == printer


def test_printed_by_property_twice(meta):
    clone = meta.clone
    printer = "Jérôme"
    printer2 = "John"
    clone.printed_by = printer
    clone.printed_by = printer2
    assert clone.printed_by == printer2


def test_get_creator(meta):
    creator = meta.creator
    expected = None
    assert creator == expected


def test_set_creator(meta):
    clone = meta.clone
    creator = "Hervé"
    clone.creator = creator
    assert clone.creator == creator


def test_set_creator_2(meta):
    clone = meta.clone
    clone.creator = None
    creator = "Hervé"
    clone.creator = creator
    assert clone.creator == creator


def test_get_keywords(meta):
    keywords = meta.get_keywords()
    expected = "Mots-clés"
    assert keywords == expected


def test_set_keywords(meta):
    clone = meta.clone
    keywords = "Nouveaux mots-clés"
    clone.set_keywords(keywords)
    assert clone.get_keywords() == keywords


def test_keyword_property(meta):
    clone = meta.clone
    keyword = "Nouveaux"
    clone.keyword = keyword
    assert clone.keyword == keyword


def test_keywords_property(meta):
    clone = meta.clone
    keywords = "Nouveaux mots-clés"
    clone.keywords = keywords
    assert clone.keywords == keywords


def test_keyword_property_on_none(meta):
    clone = meta.clone
    element = clone.get_element("//meta:keyword")
    if element:
        element.delete()
    word = "Albert"
    clone.keywords = word
    assert clone.keywords == word


def test_get_editing_duration(meta):
    duration = meta.get_editing_duration()
    expected = Duration.decode("PT00H05M30S")
    assert duration == expected


def test_set_editing_duration(meta):
    clone = meta.clone
    duration = timedelta(1, 2, 0, 0, 5, 6, 7)
    clone.set_editing_duration(duration)
    assert clone.get_editing_duration() == duration


def test_editing_duration_property(meta):
    clone = meta.clone
    duration = timedelta(1, 2, 0, 0, 5, 6, 7)
    clone.editing_duration = duration
    assert clone.editing_duration == duration
    assert clone.get_editing_duration() == duration


def test_set_bad_editing_duration(meta):
    clone = meta.clone
    duration = "PT00H01M27S"
    with pytest.raises(TypeError):
        clone.set_editing_duration(duration)


def test_get_editing_cycles(meta):
    cycles = meta.get_editing_cycles()
    expected = 2
    assert cycles == expected


def test_set_editing_cycles(meta):
    clone = meta.clone
    cycles = 1  # I swear it was a first shot!
    clone.set_editing_cycles(cycles)
    assert clone.get_editing_cycles() == cycles


def test_editing_cycles_property(meta):
    clone = meta.clone
    cycles = 1  # I swear it was a first shot!
    clone.editing_cycles = cycles
    assert clone.editing_cycles == cycles
    assert clone.get_editing_cycles() == cycles


def test_set_bad_editing_cycles(meta):
    clone = meta.clone
    cycles = "3"
    with pytest.raises(TypeError):
        clone.editing_cycles = cycles


def test_set_bad_editing_cycles_2(meta):
    clone = meta.clone
    cycles = -4
    with pytest.raises(ValueError):
        clone.editing_cycles = cycles


def test_get_generator(meta):
    generator = meta.get_generator()
    expected = (
        "LibreOffice/6.0.3.2$MacOSX_X86_64 "
        "LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821"
    )
    assert generator == expected


def test_generator_getter(meta):
    generator = meta.generator
    expected = (
        "LibreOffice/6.0.3.2$MacOSX_X86_64 "
        "LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821"
    )
    assert generator == expected


def test_set_generator(meta):
    clone = meta.clone
    generator = "odfdo Project"
    clone.set_generator(generator)
    assert clone.get_generator() == generator


def test_generator_setter(meta):
    clone = meta.clone
    generator = "odfdo Project"
    clone.generator = generator
    assert clone.generator == generator


def test_generator_default_unmodified(samples):
    document = Document(samples("meta.odt"))
    with io.BytesIO() as bytes_content:
        document.save(bytes_content)
    # document saved, generator string should be ours
    assert document.meta.get_generator() == GENERATOR


def test_generator_default_modified(samples):
    document = Document(samples("meta.odt"))
    document.meta.set_generator("custom generator")
    with io.BytesIO() as bytes_content:
        document.save(bytes_content)
    # document saved, generator string should be untouched
    assert document.meta.get_generator() == "custom generator"


def test_get_statistic(meta):
    statistic = meta.get_statistic()
    expected = {
        "meta:table-count": 0,
        "meta:image-count": 0,
        "meta:object-count": 0,
        "meta:page-count": 1,
        "meta:paragraph-count": 1,
        "meta:word-count": 4,
        "meta:character-count": 27,
        "meta:non-whitespace-character-count": 24,
    }
    assert statistic == expected


def test_set_statistic(meta):
    clone = meta.clone
    statistic = {
        "meta:table-count": 1,
        "meta:image-count": 2,
        "meta:object-count": 3,
        "meta:page-count": 4,
        "meta:paragraph-count": 5,
        "meta:word-count": 6,
        "meta:character-count": 7,
        "meta:non-whitespace-character-count": 24,
    }
    clone.set_statistic(statistic)
    assert clone.get_statistic() == statistic


def test_statistic_property(meta):
    clone = meta.clone
    statistic = {
        "meta:table-count": 1,
        "meta:image-count": 2,
        "meta:object-count": 3,
        "meta:page-count": 4,
        "meta:paragraph-count": 5,
        "meta:word-count": 6,
        "meta:character-count": 7,
        "meta:non-whitespace-character-count": 24,
    }
    clone.statistic = statistic
    assert clone.statistic == statistic
    assert clone.get_statistic() == statistic


def test_statistic_property_none(meta):
    clone = meta.clone
    element = clone.get_element("//meta:document-statistic")
    if element:
        element.delete()
    assert clone.statistic is None


def test_statistic_property_bad(meta):
    clone = meta.clone
    with pytest.raises(TypeError):
        clone.statistic = 42


def test_statistic_property_bad_2(meta):
    clone = meta.clone
    statistic = {
        "meta:table-count": "error",
        "meta:image-count": 2,
        "meta:object-count": 3,
        "meta:page-count": 4,
        "meta:paragraph-count": 5,
        "meta:word-count": 6,
        "meta:character-count": 7,
        "meta:non-whitespace-character-count": 24,
    }
    with pytest.raises(TypeError):
        clone.statistic = statistic


def test_get_user_defined_metadata(meta):
    metadata = meta.get_user_defined_metadata()
    expected = {
        "Achevé à la date": datetime(2009, 7, 31),
        "Numéro du document": Decimal("3"),
        "Référence": True,
        "Vérifié par": "Moi-même",
    }
    assert metadata == expected


def test_set_user_defined_metadata(meta):
    # A new value
    meta.set_user_defined_metadata("Prop5", 1)
    # Change a value
    meta.set_user_defined_metadata("Référence", False)
    expected = {
        "Achevé à la date": datetime(2009, 7, 31),
        "Numéro du document": Decimal("3"),
        "Référence": False,
        "Vérifié par": "Moi-même",
        "Prop5": Decimal("1"),
    }
    metadata = meta.get_user_defined_metadata()
    assert metadata == expected


def test_user_defined_metadata_property_getter(meta):
    # A new value
    meta.set_user_defined_metadata("Prop5", 1)
    # Change a value
    meta.set_user_defined_metadata("Référence", False)
    expected = {
        "Achevé à la date": datetime(2009, 7, 31),
        "Numéro du document": Decimal("3"),
        "Référence": False,
        "Vérifié par": "Moi-même",
        "Prop5": Decimal("1"),
    }
    metadata = meta.user_defined_metadata
    assert metadata == expected


def test_set_user_defined_metadata_none(meta):
    # A new value
    meta.set_user_defined_metadata("Prop5", 2)
    # Change a value
    meta.set_user_defined_metadata("Référence", None)
    meta.set_user_defined_metadata("Vérifié par", None)
    expected = {
        "Achevé à la date": datetime(2009, 7, 31),
        "Numéro du document": Decimal("3"),
        "Prop5": Decimal("2"),
    }
    metadata = meta.get_user_defined_metadata()
    assert metadata == expected


def test_user_defined_metadata_clear(meta):
    meta.clear_user_defined_metadata()
    metadata = meta.user_defined_metadata
    assert metadata == {}


def test_user_defined_metadata_property_setter(meta):
    new_data = {
        "Achevé le date": datetime(2012, 7, 31),
        "Numéro document": Decimal("4"),
        "Référence": True,
    }
    meta.user_defined_metadata = new_data
    assert meta.user_defined_metadata == new_data
    assert meta.get_user_defined_metadata() == new_data


def test_set_user_defined_metadata_datetime(meta):
    dt = datetime(2025, 1, 2, 3, 4, 5)
    expected = {
        "name": "when",
        "text": "2025-01-02T03:04:05",
        "value": dt,
        "value_type": "date",
    }
    meta.set_user_defined_metadata("when", dt)
    assert meta.get_user_defined_metadata_of_name("when") == expected


def test_set_user_defined_metadata_date(meta):
    dt = dtdate(2025, 1, 2)
    expected = {
        "name": "when",
        "text": "2025-01-02",
        "value": datetime(2025, 1, 2, 0, 0, 0),
        "value_type": "date",
    }
    meta.set_user_defined_metadata("when", dt)
    assert meta.get_user_defined_metadata_of_name("when") == expected


def test_set_user_defined_metadata_timedelta(meta):
    td = timedelta(hours=4)
    expected = {
        "name": "when",
        "text": "PT04H00M00S",
        "value": td,
        "value_type": "time",
    }
    meta.set_user_defined_metadata("when", td)
    assert meta.get_user_defined_metadata_of_name("when") == expected


def test_set_user_defined_metadata_wrong_type(meta):
    with pytest.raises(TypeError):
        meta.set_user_defined_metadata("when", [])


def test_get_user_defined_metadata_of_name(meta):
    ref = "Référence"
    metadata = meta.get_user_defined_metadata_of_name(ref)
    expected = {"name": ref, "text": "true", "value": True, "value_type": "boolean"}
    assert metadata == expected


def test_get_user_defined_metadata_buggy_meta_name(meta):
    ref = "Référence"
    for item in meta.get_elements("//meta:user-defined"):
        # Read the values
        name = item.get_attribute("meta:name")
        if name == ref:
            item.set_attribute("meta:name", None)
            break
    metadata = meta.get_user_defined_metadata()
    expected = {
        "Achevé à la date": datetime(2009, 7, 31),
        "Numéro du document": Decimal("3"),
        "Vérifié par": "Moi-même",
        "": True,
    }
    assert metadata == expected


def test_user_defined_metadata_list_buggy_meta_name(meta):
    ref = "Référence"
    for item in meta.get_elements("//meta:user-defined"):
        # Read the values
        name = item.get_attribute("meta:name")
        if name == ref:
            item.set_attribute("meta:name", None)
            break
    metadata_list = meta._user_defined_metadata_list()
    expected = [
        {
            "meta:name": "",
            "meta:value-type": "boolean",
            "value": True,
        },
        {
            "meta:name": "Achevé à la date",
            "meta:value-type": "date",
            "value": datetime(2009, 7, 31, 0, 0),
        },
        {
            "meta:name": "Numéro du document",
            "meta:value-type": "float",
            "value": Decimal("3"),
        },
        {
            "meta:name": "Vérifié par",
            "meta:value-type": "string",
            "value": "Moi-même",
        },
    ]
    assert metadata_list == expected


def test_user_defined_metadata_list(meta):
    td = timedelta(hours=5)
    meta.set_user_defined_metadata("duration", td)
    metadata_list = meta._user_defined_metadata_list()
    expected = [
        {
            "meta:name": "Achevé à la date",
            "meta:value-type": "date",
            "value": datetime(2009, 7, 31, 0, 0),
        },
        {
            "meta:name": "Numéro du document",
            "meta:value-type": "float",
            "value": Decimal("3"),
        },
        {"meta:name": "Référence", "meta:value-type": "boolean", "value": True},
        {"meta:name": "Vérifié par", "meta:value-type": "string", "value": "Moi-même"},
        {
            "meta:name": "duration",
            "meta:value-type": "time",
            "value": td,
        },
    ]
    assert metadata_list == expected


def test_no_meta_template(meta):
    clone = meta.clone
    template = clone.template
    assert template is None


def test_set_meta_template(meta):
    clone = meta.clone
    current = clone.template
    if isinstance(current, MetaTemplate):
        current.delete()
    now = datetime.now().replace(microsecond=0)
    clone.set_template(date=now, href="some url", title="some title")
    template = clone.template
    assert isinstance(template, MetaTemplate)
    assert template.date == DateTime.encode(now)
    assert template.href == "some url"
    assert template.title == "some title"


def test_set_meta_template_twice(meta):
    clone = meta.clone
    now = datetime.now().replace(microsecond=0)
    clone.set_template(date=now, href="some url", title="some title")
    template = clone.template
    assert isinstance(template, MetaTemplate)
    # second change
    clone.set_template(date=now, href="new url", title="new title")
    template = clone.template
    assert isinstance(template, MetaTemplate)
    assert template.date == DateTime.encode(now)
    assert template.href == "new url"
    assert template.title == "new title"


def test_no_meta_auto_reload(meta):
    clone = meta.clone
    reload = clone.auto_reload
    assert reload is None


def test_set_meta_auto_reload(meta):
    clone = meta.clone
    current = clone.auto_reload
    if isinstance(current, MetaAutoReload):
        current.delete()
    delay = timedelta(seconds=15)
    clone.set_auto_reload(delay=delay, href="some url")
    reload = clone.auto_reload
    assert isinstance(reload, MetaAutoReload)
    assert reload.delay == Duration.encode(delay)
    assert reload.href == "some url"
    assert repr(reload) == (
        "<MetaAutoReload tag=meta:auto-reload href=some url delay=0:00:15>"
    )


def test_set_meta_auto_reload_twice(meta):
    clone = meta.clone
    delay = timedelta(seconds=15)
    clone.set_auto_reload(delay=delay, href="some url")
    reload = clone.auto_reload
    assert isinstance(reload, MetaAutoReload)
    # second change
    clone.set_auto_reload(delay=delay, href="new url")
    reload = clone.auto_reload
    assert isinstance(reload, MetaAutoReload)
    assert reload.delay == Duration.encode(delay)
    assert reload.href == "new url"
    assert repr(reload) == (
        "<MetaAutoReload tag=meta:auto-reload href=new url delay=0:00:15>"
    )


def test_no_hyperlink_behaviour(meta):
    clone = meta.clone
    behaviour = clone.hyperlink_behaviour
    assert behaviour is None


def test_set_hyperlink_behaviour(meta):
    clone = meta.clone
    current = clone.hyperlink_behaviour
    if isinstance(current, MetaHyperlinkBehaviour):
        current.delete()
    clone.set_hyperlink_behaviour(target_frame_name="some_frame", show="_top")
    behaviour = clone.hyperlink_behaviour
    assert isinstance(behaviour, MetaHyperlinkBehaviour)
    assert behaviour.show == "_top"
    assert behaviour.target_frame_name == "some_frame"
    assert repr(behaviour) == (
        "<MetaHyperlinkBehaviour tag=meta:hyperlink-behaviour target=some_frame show=_top>"
    )


def test_set_hyperlink_behaviour_twice(meta):
    clone = meta.clone
    clone.set_hyperlink_behaviour(target_frame_name="some_frame", show="_top")
    behaviour = clone.hyperlink_behaviour
    assert isinstance(behaviour, MetaHyperlinkBehaviour)
    # second change
    clone.set_hyperlink_behaviour(target_frame_name="new_frame", show="_top")
    behaviour = clone.hyperlink_behaviour
    assert isinstance(behaviour, MetaHyperlinkBehaviour)
    assert behaviour.show == "_top"
    assert behaviour.target_frame_name == "new_frame"
    assert repr(behaviour) == (
        "<MetaHyperlinkBehaviour tag=meta:hyperlink-behaviour target=new_frame show=_top>"
    )


def test_meta_export_dict(meta):
    exported = meta.as_dict()
    expected = {
        "meta:creation-date": datetime(2009, 7, 31, 15, 57, 37),
        "dc:date": datetime(2009, 7, 31, 15, 59, 13),
        "meta:editing-duration": timedelta(seconds=330),
        "meta:editing-cycles": 2,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 1,
            "meta:paragraph-count": 1,
            "meta:word-count": 4,
            "meta:character-count": 27,
            "meta:non-whitespace-character-count": 24,
        },
        "meta:generator": "LibreOffice/6.0.3.2$MacOSX_X86_64 LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821",
        "dc:title": "Intitulé",
        "dc:description": "Comments\nCommentaires\n评论",
        "meta:keyword": "Mots-clés",
        "dc:subject": "Sujet de sa majesté",
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "meta:value-type": "date",
                "value": datetime(2009, 7, 31, 0, 0),
            },
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": Decimal("3"),
            },
            {"meta:name": "Référence", "meta:value-type": "boolean", "value": True},
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même",
            },
        ],
    }
    assert exported == expected


def test_meta_statistic_none(meta):
    with pytest.raises(TypeError):
        meta.statistic = None


def test_meta_export_dict_error(meta):
    element = meta.get_element("//meta:document-statistic")
    element.delete()
    with pytest.raises(LookupError):
        meta.as_dict()


def test_meta_export_dict_auto_reload_hyperlink_behaviour(meta):
    meta.set_auto_reload(timedelta(seconds=30))
    meta.set_hyperlink_behaviour()
    exported = meta.as_dict()
    expected = {
        "meta:creation-date": datetime(2009, 7, 31, 15, 57, 37),
        "dc:date": datetime(2009, 7, 31, 15, 59, 13),
        "meta:editing-duration": timedelta(seconds=330),
        "meta:editing-cycles": 2,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 1,
            "meta:paragraph-count": 1,
            "meta:word-count": 4,
            "meta:character-count": 27,
            "meta:non-whitespace-character-count": 24,
        },
        "meta:generator": "LibreOffice/6.0.3.2$MacOSX_X86_64 LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821",
        "dc:title": "Intitulé",
        "dc:description": "Comments\nCommentaires\n评论",
        "meta:keyword": "Mots-clés",
        "dc:subject": "Sujet de sa majesté",
        "meta:auto-reload": {
            "meta:delay": timedelta(seconds=30),
            "xlink:actuate": "onLoad",
            "xlink:href": "",
            "xlink:show": "replace",
            "xlink:type": "simple",
        },
        "meta:hyperlink-behaviour": {
            "office:target-frame-name": "_blank",
            "xlink:show": "replace",
        },
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "meta:value-type": "date",
                "value": datetime(2009, 7, 31, 0, 0),
            },
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": Decimal("3"),
            },
            {"meta:name": "Référence", "meta:value-type": "boolean", "value": True},
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même",
            },
        ],
    }
    assert exported == expected


def test_meta_export_dict_full(meta):
    exported = meta.as_dict(True)
    expected = {
        "meta:creation-date": datetime(2009, 7, 31, 15, 57, 37),
        "dc:date": datetime(2009, 7, 31, 15, 59, 13),
        "meta:editing-duration": timedelta(seconds=330),
        "meta:editing-cycles": 2,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 1,
            "meta:paragraph-count": 1,
            "meta:word-count": 4,
            "meta:character-count": 27,
            "meta:non-whitespace-character-count": 24,
        },
        "meta:generator": "LibreOffice/6.0.3.2$MacOSX_X86_64 LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821",
        "dc:title": "Intitulé",
        "dc:description": "Comments\nCommentaires\n评论",
        "dc:creator": None,
        "meta:keyword": "Mots-clés",
        "dc:subject": "Sujet de sa majesté",
        "dc:language": None,
        "meta:initial-creator": None,
        "meta:print-date": None,
        "meta:printed-by": None,
        "meta:auto-reload": None,
        "meta:hyperlink-behaviour": None,
        "meta:template": None,
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "meta:value-type": "date",
                "value": datetime(2009, 7, 31, 0, 0),
            },
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": Decimal("3"),
            },
            {"meta:name": "Référence", "meta:value-type": "boolean", "value": True},
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même",
            },
        ],
    }
    assert exported == expected


def test_meta_export_json(meta):
    exported = meta.as_json()
    expected = dedent(
        """\
    {
        "meta:creation-date": "2009-07-31T15:57:37",
        "dc:date": "2009-07-31T15:59:13",
        "meta:editing-duration": "PT00H05M30S",
        "meta:editing-cycles": 2,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 1,
            "meta:paragraph-count": 1,
            "meta:word-count": 4,
            "meta:character-count": 27,
            "meta:non-whitespace-character-count": 24
        },
        "meta:generator": "LibreOffice/6.0.3.2$MacOSX_X86_64 LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821",
        "dc:title": "Intitulé",
        "dc:description": "Comments\\nCommentaires\\n评论",
        "meta:keyword": "Mots-clés",
        "dc:subject": "Sujet de sa majesté",
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "meta:value-type": "date",
                "value": "2009-07-31T00:00:00"
            },
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": 3
            },
            {
                "meta:name": "Référence",
                "meta:value-type": "boolean",
                "value": true
            },
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même"
            }
        ]
    }"""
    )
    assert exported.strip() == expected.strip()


def test_meta_export_json_with_reload(meta):
    delay = timedelta(seconds=15)
    meta.set_auto_reload(delay, "some_url")
    exported = meta.as_json()
    expected = dedent(
        """\
    {
        "meta:creation-date": "2009-07-31T15:57:37",
        "dc:date": "2009-07-31T15:59:13",
        "meta:editing-duration": "PT00H05M30S",
        "meta:editing-cycles": 2,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 1,
            "meta:paragraph-count": 1,
            "meta:word-count": 4,
            "meta:character-count": 27,
            "meta:non-whitespace-character-count": 24
        },
        "meta:generator": "LibreOffice/6.0.3.2$MacOSX_X86_64 LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821",
        "dc:title": "Intitulé",
        "dc:description": "Comments\\nCommentaires\\n评论",
        "meta:keyword": "Mots-clés",
        "dc:subject": "Sujet de sa majesté",
        "meta:auto-reload": {
            "meta:delay": "PT00H00M15S",
            "xlink:actuate": "onLoad",
            "xlink:href": "some_url",
            "xlink:show": "replace",
            "xlink:type": "simple"
        },
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "meta:value-type": "date",
                "value": "2009-07-31T00:00:00"
            },
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": 3
            },
            {
                "meta:name": "Référence",
                "meta:value-type": "boolean",
                "value": true
            },
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même"
            }
        ]
    }"""
    )
    assert exported.strip() == expected.strip()


def test_meta_export_json_full(meta):
    exported = meta.as_json(True)
    expected = dedent(
        """\
    {
        "meta:creation-date": "2009-07-31T15:57:37",
        "dc:date": "2009-07-31T15:59:13",
        "meta:editing-duration": "PT00H05M30S",
        "meta:editing-cycles": 2,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 1,
            "meta:paragraph-count": 1,
            "meta:word-count": 4,
            "meta:character-count": 27,
            "meta:non-whitespace-character-count": 24
        },
        "meta:generator": "LibreOffice/6.0.3.2$MacOSX_X86_64 LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821",
        "dc:title": "Intitulé",
        "dc:description": "Comments\\nCommentaires\\n评论",
        "dc:creator": null,
        "meta:keyword": "Mots-clés",
        "dc:subject": "Sujet de sa majesté",
        "dc:language": null,
        "meta:initial-creator": null,
        "meta:print-date": null,
        "meta:printed-by": null,
        "meta:template": null,
        "meta:auto-reload": null,
        "meta:hyperlink-behaviour": null,
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "meta:value-type": "date",
                "value": "2009-07-31T00:00:00"
            },
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": 3
            },
            {
                "meta:name": "Référence",
                "meta:value-type": "boolean",
                "value": true
            },
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même"
            }
        ]
    }"""
    )
    assert exported == expected


def test_meta_from_dict_1(meta):
    imported = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2009, 7, 31, 15, 59, 13),
        "meta:editing-duration": timedelta(seconds=100),
        "meta:editing-cycles": 1,
        "dc:subject": None,
    }
    expected = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2024, 7, 14, 12, 00, 00),
        "meta:editing-duration": timedelta(seconds=0),
        "meta:editing-cycles": 1,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 1,
            "meta:paragraph-count": 1,
            "meta:word-count": 4,
            "meta:character-count": 27,
            "meta:non-whitespace-character-count": 24,
        },
        "meta:generator": "LibreOffice/6.0.3.2$MacOSX_X86_64 LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821",
        "dc:title": "Intitulé",
        "dc:description": "Comments\nCommentaires\n评论",
        "dc:creator": None,
        "meta:keyword": "Mots-clés",
        "dc:subject": None,
        "dc:language": None,
        "meta:initial-creator": None,
        "meta:print-date": None,
        "meta:printed-by": None,
        "meta:auto-reload": None,
        "meta:hyperlink-behaviour": None,
        "meta:template": None,
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "meta:value-type": "date",
                "value": datetime(2009, 7, 31, 0, 0),
            },
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": Decimal("3"),
            },
            {"meta:name": "Référence", "meta:value-type": "boolean", "value": True},
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même",
            },
        ],
    }
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result == expected


def test_meta_from_dict_2(meta):
    imported = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2009, 7, 31, 15, 59, 13),
        "meta:editing-duration": timedelta(seconds=100),
        "meta:editing-cycles": 1,
        "dc:subject": None,
        "meta:user-defined": None,
    }
    expected = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2024, 7, 14, 12, 00, 00),
        "meta:editing-duration": timedelta(seconds=0),
        "meta:editing-cycles": 1,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 1,
            "meta:paragraph-count": 1,
            "meta:word-count": 4,
            "meta:character-count": 27,
            "meta:non-whitespace-character-count": 24,
        },
        "meta:generator": "LibreOffice/6.0.3.2$MacOSX_X86_64 LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821",
        "dc:title": "Intitulé",
        "dc:description": "Comments\nCommentaires\n评论",
        "dc:creator": None,
        "meta:keyword": "Mots-clés",
        "dc:subject": None,
        "dc:language": None,
        "meta:initial-creator": None,
        "meta:print-date": None,
        "meta:printed-by": None,
        "meta:auto-reload": None,
        "meta:hyperlink-behaviour": None,
        "meta:template": None,
        "meta:user-defined": [],
    }
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result == expected


def test_meta_from_dict_3(meta):
    imported = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2009, 7, 31, 15, 59, 13),
        "meta:editing-duration": timedelta(seconds=100),
        "meta:editing-cycles": 1,
        "dc:subject": None,
        "meta:user-defined": [
            {
                "meta:name": "New property",
                "meta:value-type": "string",
                "value": "Anatole",
            }
        ],
    }
    expected = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2024, 7, 14, 12, 00, 00),
        "meta:editing-duration": timedelta(seconds=0),
        "meta:editing-cycles": 1,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 1,
            "meta:paragraph-count": 1,
            "meta:word-count": 4,
            "meta:character-count": 27,
            "meta:non-whitespace-character-count": 24,
        },
        "meta:generator": "LibreOffice/6.0.3.2$MacOSX_X86_64 LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821",
        "dc:title": "Intitulé",
        "dc:description": "Comments\nCommentaires\n评论",
        "dc:creator": None,
        "meta:keyword": "Mots-clés",
        "dc:subject": None,
        "dc:language": None,
        "meta:initial-creator": None,
        "meta:print-date": None,
        "meta:printed-by": None,
        "meta:auto-reload": None,
        "meta:hyperlink-behaviour": None,
        "meta:template": None,
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "meta:value-type": "date",
                "value": datetime(2009, 7, 31, 0, 0),
            },
            {
                "meta:name": "New property",
                "meta:value-type": "string",
                "value": "Anatole",
            },
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": Decimal("3"),
            },
            {"meta:name": "Référence", "meta:value-type": "boolean", "value": True},
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même",
            },
        ],
    }
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result == expected


def test_meta_from_dict_4(meta):
    imported = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2009, 7, 31, 15, 59, 13),
        "meta:editing-duration": timedelta(seconds=100),
        "meta:editing-cycles": 1,
        "meta:document-statistic": None,
        "meta:generator": "toto",
        "dc:subject": None,
        "dc:title": None,
        "dc:description": None,
        "dc:keyword": None,
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "value": None,
            }
        ],
    }
    expected = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2024, 7, 14, 12, 00, 00),
        "meta:editing-duration": timedelta(seconds=0),
        "meta:editing-cycles": 1,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 0,
            "meta:paragraph-count": 0,
            "meta:word-count": 0,
            "meta:character-count": 0,
            "meta:non-whitespace-character-count": 0,
        },
        "meta:generator": "toto",
        "dc:title": None,
        "dc:description": None,
        "dc:creator": None,
        "meta:keyword": "Mots-clés",
        "dc:subject": None,
        "dc:language": None,
        "meta:initial-creator": None,
        "meta:print-date": None,
        "meta:printed-by": None,
        "meta:auto-reload": None,
        "meta:hyperlink-behaviour": None,
        "meta:template": None,
        "meta:user-defined": [
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": Decimal("3"),
            },
            {"meta:name": "Référence", "meta:value-type": "boolean", "value": True},
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même",
            },
        ],
    }
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result == expected


def test_meta_from_dict_5(meta):
    imported = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2009, 7, 31, 15, 59, 13),
        "meta:editing-duration": timedelta(seconds=100),
        "meta:editing-cycles": 1,
        "meta:document-statistic": {
            "meta:table-count": 1,
            "meta:image-count": 22,
            "meta:object-count": 33,
            "meta:page-count": 444,
            "meta:paragraph-count": 555,
            "meta:word-count": 666,
            "meta:character-count": 777,
            "meta:non-whitespace-character-count": 888,
        },
        "meta:generator": "toto",
        "dc:subject": None,
        "dc:title": None,
        "dc:description": None,
        "dc:keyword": None,
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "value": None,
            }
        ],
    }
    expected = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2024, 7, 14, 12, 00, 00),
        "meta:editing-cycles": 1,
        "meta:document-statistic": {
            "meta:table-count": 1,
            "meta:image-count": 22,
            "meta:object-count": 33,
            "meta:page-count": 444,
            "meta:paragraph-count": 555,
            "meta:word-count": 666,
            "meta:character-count": 777,
            "meta:non-whitespace-character-count": 888,
        },
        "meta:generator": "toto",
        "meta:keyword": "Mots-clés",
        "meta:user-defined": [
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": Decimal("3"),
            },
            {"meta:name": "Référence", "meta:value-type": "boolean", "value": True},
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même",
            },
        ],
    }
    meta.from_dict(imported)
    result = meta.as_dict(full=False)
    assert result == expected


def test_meta_from_dict_42(meta):
    imported = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2009, 7, 31, 15, 59, 13),
        "meta:editing-duration": timedelta(seconds=100),
        "meta:editing-cycles": 1,
        "meta:document-statistic": None,
        "meta:generator": "toto",
        "dc:subject": None,
        "dc:title": None,
        "dc:description": None,
        "dc:keyword": None,
        "meta:user-defined": [
            {
                "meta:name": "Achevé à la date",
                "value": None,
            }
        ],
    }
    expected = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 00, 00),
        "dc:date": datetime(2024, 7, 14, 12, 00, 00),
        "meta:editing-duration": timedelta(seconds=0),
        "meta:editing-cycles": 1,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 0,
            "meta:paragraph-count": 0,
            "meta:word-count": 0,
            "meta:character-count": 0,
            "meta:non-whitespace-character-count": 0,
        },
        "meta:generator": "toto",
        "dc:title": None,
        "dc:description": None,
        "dc:creator": None,
        "meta:keyword": "Mots-clés",
        "dc:subject": None,
        "dc:language": None,
        "meta:initial-creator": None,
        "meta:print-date": None,
        "meta:printed-by": None,
        "meta:auto-reload": None,
        "meta:hyperlink-behaviour": None,
        "meta:template": None,
        "meta:user-defined": [
            {
                "meta:name": "Numéro du document",
                "meta:value-type": "float",
                "value": Decimal("3"),
            },
            {"meta:name": "Référence", "meta:value-type": "boolean", "value": True},
            {
                "meta:name": "Vérifié par",
                "meta:value-type": "string",
                "value": "Moi-même",
            },
        ],
    }
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result == expected


def test_meta_from_dict_modif_date(meta):
    imported = {}
    element = meta.get_element("//dc:date")
    element.delete()
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert isinstance(result["dc:date"], datetime)


def test_meta_from_dict_creation_date(meta):
    imported = {}
    element = meta.get_element("//meta:creation-date")
    element.delete()
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert isinstance(result["meta:creation-date"], datetime)


def test_meta_from_dict_generator(meta):
    imported = {}
    meta.generator = ""
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result["meta:generator"].startswith("odfdo")


def test_meta_from_dict_creator(meta):
    key = "dc:creator"
    content = "John"
    imported = {key: content}
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result[key] == content


def test_meta_from_dict_language(meta):
    key = "dc:language"
    content = "fr-FR"
    imported = {key: content}
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result[key] == content


def test_meta_from_dict_initial_creator(meta):
    key = "meta:initial-creator"
    content = "Homer"
    imported = {key: content}
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result[key] == content


def test_meta_from_dict_printed_by(meta):
    key = "meta:printed-by"
    content = "Homer"
    imported = {key: content}
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result[key] == content


def test_meta_from_dict_print_date(meta):
    key = "meta:print-date"
    content = datetime(2025, 1, 2, 3, 4, 5, 6)
    imported = {key: content}
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result[key] == content


def test_meta_from_dict_meta_template(meta):
    key = "meta:template"
    template = MetaTemplate(datetime(2025, 1, 2, 3, 4, 5, 6), "url", "Title")
    content = template.as_dict()
    imported = {key: content}
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    assert result[key] == content


def test_meta_from_dict_auto_reload(meta):
    key = "meta:auto-reload"
    delay = timedelta(seconds=15)
    meta.set_auto_reload(delay=delay, href="new url")
    imported = {}
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    content = {
        "meta:delay": delay,
        "xlink:actuate": "onLoad",
        "xlink:href": "new url",
        "xlink:show": "replace",
        "xlink:type": "simple",
    }
    assert result[key] == content


def test_meta_from_dict_behaviour(meta):
    key = "meta:hyperlink-behaviour"
    meta.set_hyperlink_behaviour(target_frame_name="some_frame", show="_top")
    imported = {}
    meta.from_dict(imported)
    result = meta.as_dict(full=True)
    content = {
        "office:target-frame-name": "some_frame",
        "xlink:show": "_top",
    }
    assert result[key] == content


def test_meta_strip_1(samples):
    document = Document(samples("meta.odt"))
    meta = document.meta
    dt = datetime(2024, 7, 14, 12, 13, 14)
    meta.strip(creation_date=dt)
    # check document updated
    meta = document.meta
    exported = meta.as_dict(True)
    expected = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 13, 14),
        "dc:date": datetime(2024, 7, 14, 12, 13, 14),
        "meta:editing-duration": timedelta(0),
        "meta:editing-cycles": 1,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 0,
            "meta:paragraph-count": 0,
            "meta:word-count": 0,
            "meta:character-count": 0,
            "meta:non-whitespace-character-count": 0,
        },
        "meta:generator": GENERATOR,
        "dc:title": None,
        "dc:description": None,
        "dc:creator": None,
        "meta:keyword": None,
        "dc:subject": None,
        "dc:language": None,
        "meta:initial-creator": None,
        "meta:print-date": None,
        "meta:printed-by": None,
        "meta:template": None,
        "meta:auto-reload": None,
        "meta:hyperlink-behaviour": None,
        "meta:user-defined": [],
    }

    assert exported == expected


def test_meta_strip_2(samples):
    document = Document(samples("meta.odt"))
    meta = document.meta
    dt = datetime(2024, 7, 14, 12, 13, 14)
    meta.strip(generator="unknown", creation_date=dt)
    # check document updated
    meta = document.meta
    exported = meta.as_dict(True)
    expected = {
        "meta:creation-date": datetime(2024, 7, 14, 12, 13, 14),
        "dc:date": datetime(2024, 7, 14, 12, 13, 14),
        "meta:editing-duration": timedelta(0),
        "meta:editing-cycles": 1,
        "meta:document-statistic": {
            "meta:table-count": 0,
            "meta:image-count": 0,
            "meta:object-count": 0,
            "meta:page-count": 0,
            "meta:paragraph-count": 0,
            "meta:word-count": 0,
            "meta:character-count": 0,
            "meta:non-whitespace-character-count": 0,
        },
        "meta:generator": "unknown",
        "dc:title": None,
        "dc:description": None,
        "dc:creator": None,
        "meta:keyword": None,
        "dc:subject": None,
        "dc:language": None,
        "meta:initial-creator": None,
        "meta:print-date": None,
        "meta:printed-by": None,
        "meta:template": None,
        "meta:auto-reload": None,
        "meta:hyperlink-behaviour": None,
        "meta:user-defined": [],
    }

    assert exported == expected


def test_meta_body_class(meta):
    body = meta.body
    assert isinstance(body, Metadata)


def test_meta_body_serialize(meta):
    body = meta.body
    expected = (
        "<office:meta>"
        "<meta:creation-date>2009-07-31T15:57:37</meta:creation-date>"
        "<meta:editing-duration>PT00H05M30S</meta:editing-duration>"
        "<meta:editing-cycles>2</meta:editing-cycles>"
        "<meta:generator>LibreOffice/6.0.3.2$MacOSX_X86_64 "
        "LibreOffice_project/8f48d515416608e3a835360314dac7e47fd0b821"
        "</meta:generator>"
        "<dc:description>Comments\nCommentaires\n评论</dc:description>"
        "<meta:keyword>Mots-clés</meta:keyword>"
        "<dc:subject>Sujet de sa majesté</dc:subject>"
        "<dc:title>Intitulé</dc:title>"
        "<dc:date>2009-07-31T15:59:13</dc:date>"
        '<meta:document-statistic meta:table-count="0" '
        'meta:image-count="0" meta:object-count="0" '
        'meta:page-count="1" meta:paragraph-count="1" '
        'meta:word-count="4" meta:character-count="27" '
        'meta:non-whitespace-character-count="24"/>'
        '<meta:user-defined meta:name="Achevé à la date" '
        'meta:value-type="date">2009-07-31T00:00:00</meta:user-defined>'
        '<meta:user-defined meta:name="Numéro du document" '
        'meta:value-type="float">3</meta:user-defined>'
        '<meta:user-defined meta:name="Référence" '
        'meta:value-type="boolean">true</meta:user-defined>'
        '<meta:user-defined meta:name="Vérifié par">Moi-même'
        "</meta:user-defined>"
        "</office:meta>"
    )
    assert body.serialize() == expected


def test_meta_body_clear(meta):
    body = meta.body
    body.clear()
    expected = "<office:meta/>"
    assert body.serialize() == expected


def test_meta_body_from_no_dict_1(meta):
    meta2 = meta.clone
    body2 = meta2.body
    body2.clear()
    meta2.set_statistic({})
    expected = (
        "{\n"
        '    "meta:document-statistic": {\n'
        '        "meta:table-count": 0,\n'
        '        "meta:image-count": 0,\n'
        '        "meta:object-count": 0,\n'
        '        "meta:page-count": 0,\n'
        '        "meta:paragraph-count": 0,\n'
        '        "meta:word-count": 0,\n'
        '        "meta:character-count": 0,\n'
        '        "meta:non-whitespace-character-count": 0\n'
        "    },\n"
        '    "meta:user-defined": []\n'
        "}"
    )
    assert meta2.as_json() == expected


def test_meta_body_from_no_dict_2(meta):
    meta2 = meta.clone
    meta.body.clear()
    meta.set_statistic({})
    body2 = meta2.body
    body2.clear()
    meta2.set_statistic({})
    meta2_dict = meta2.as_dict()
    meta.date = datetime(2025, 1, 2, 3, 4, 5, 6).replace(microsecond=0)
    meta.creation_date = datetime(2025, 1, 2, 3, 4, 5, 6).replace(microsecond=0)
    meta.generator = "gen"
    meta.from_dict(meta2_dict)
    expected = (
        "{\n"
        '    "meta:creation-date": "2025-01-02T03:04:05",\n'
        '    "dc:date": "2025-01-02T03:04:05",\n'
        '    "meta:editing-cycles": 1,\n'
        '    "meta:document-statistic": {\n'
        '        "meta:table-count": 0,\n'
        '        "meta:image-count": 0,\n'
        '        "meta:object-count": 0,\n'
        '        "meta:page-count": 0,\n'
        '        "meta:paragraph-count": 0,\n'
        '        "meta:word-count": 0,\n'
        '        "meta:character-count": 0,\n'
        '        "meta:non-whitespace-character-count": 0\n'
        "    },\n"
        '    "meta:generator": "gen",\n'
        '    "meta:user-defined": []\n'
        "}"
    )
    assert meta.as_json() == expected


def test_deprecated_meta_body(meta):
    assert meta.meta_body.serialize() == meta.body.serialize()



#!/usr/bin/env python
"""Open a table of 1000 lines and 100 columns, extract a sub table
of 100 lines 26 columns, save the result in a spreadsheet document.
"""

import os
import sys
from pathlib import Path

from odfdo import Document, Row, Table

_DOC_SEQUENCE = 450
DATA = Path(__file__).parent / "data"
SOURCE = "big_table.ods"
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "extract_table"
TARGET = "document.ods"


def read_source_document() -> Document:
    """Return the source Document."""
    try:
        source = sys.argv[1]
    except IndexError:
        source = DATA / SOURCE
    return Document(source)


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


def extract_sub_tables(document: Document) -> Document:
    """Return a spreadsheet with 2 sheets extracted from the imput document."""
    # Expected_result:
    #  Size of Big Table : (100, 1000)
    #  Size of extracted table 1 : (26, 100)
    #  Size of extracted table 2 : (26, 100)
    #
    big_table = document.body.get_table(name="Big Table")
    print("Size of Big Table :", big_table.size)

    extracted = Document("ods")
    extracted.body.clear()
    # now extract 100 rows of 26 columns :
    table1 = Table("Extract 1")
    for r in range(800, 900):
        row = big_table.get_row(r)
        extracted_values = [row.get_value(x) for x in range(50, 76)]
        new_row = Row()
        new_row.set_values(extracted_values)
        table1.append(new_row)
    extracted.body.append(table1)
    print("Size of extracted table 1 :", table1.size)

    # other method
    table2 = Table("Extract 2")
    cells = big_table.get_cells(coord=(50, 800, 75, 899))
    table2.set_cells(coord=(0, 0), cells=cells)
    extracted.body.append(table2)
    print("Size of extracted table 2 :", table2.size)

    return extracted


def main() -> None:
    document = read_source_document()
    extracted = extract_sub_tables(document)
    test_unit(extracted)
    save_new(extracted, TARGET)


def test_unit(spreadsheet: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    body = spreadsheet.body
    table1 = body.get_table(position=0)
    assert table1.size == (26, 100)
    table2 = body.get_table(position=1)
    assert table2.size == (26, 100)


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Classes for ODF document user fields declarations.

This module provides classes for managing user field declarations and their
containers within the document content, tags "text:user-field-decls",
"text:user-field-decl".
.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union, cast

from .element import FIRST_CHILD, Element, PropDef, register_element_class
from .element_typed import ElementTyped

if TYPE_CHECKING:
    from datetime import datetime, timedelta
    from decimal import Decimal


class UserFieldDeclMixin(Element):
    """Mixin class for elements that can contain user field declarations.

    This mixin provides methods to access "text:user-field-decl".

    Used by the following classes:
        - "office:chart"
        - "office:drawing"
        - "office:presentation"
        - "office:spreadsheet"
        - "office:text"
        - "style:footer"
        - "style:footer-first"
        - "style:footer-left"
        - "style:header"
        - "style:header-first"
        - "style:header-left"
        - "text:user-field-decls"
    """

    def get_user_field_decl_list(self) -> list[UserFieldDecl]:
        """Returns all user field declarations as a list.

        Returns:
            list[UserFieldDecl]: A list of all UserFieldDecl instances that
            are descendants of this element.
        """
        return cast(
            list[UserFieldDecl],
            self._filtered_elements(
                "descendant::text:user-field-decl",
            ),
        )

    def get_user_field_decl(self, name: str, position: int = 0) -> UserFieldDecl | None:
        """Returns a single user field declaration that matches the specified
        criteria.

        Args:
            name: The name of the user field declaration to retrieve.
            position: The 0-based index of the matching user field
                declaration to return.

        Returns:
            UserFieldDecl | None: A UserFieldDecl instance, or None if no
            declaration matches the criteria.
        """
        return cast(
            Union[None, UserFieldDecl],
            self._filtered_element(
                "descendant::text:user-field-decl", position, text_name=name
            ),
        )

    def get_user_field_value(
        self, name: str, value_type: str | None = None
    ) -> bool | str | int | float | Decimal | datetime | timedelta | None:
        """Returns the value of the specified user field.

        Args:
            name: The name of the user field to retrieve its value.
            value_type: The expected type of the user field's value.
                Can be 'boolean', 'currency', 'date', 'float',
                'percentage', 'string', 'time', or None for automatic
                type detection.

        Returns:
            bool | str | int | float | Decimal | datetime | timedelta | None:
                The value of the user field, cast to the most appropriate
                Python type, or None if the user field is not found.
        """
        user_field_decl = self.get_user_field_decl(name)
        if user_field_decl is None:
            return None
        value = user_field_decl.get_value(value_type)
        return value  # type: ignore


class UserFieldDeclContMixin(UserFieldDeclMixin):
    """Mixin class for elements that can contain user field declarations.

    This mixin provides methods to access and manipulate
    "text:user-field-decls" and "text:user-field-decl".

    Used by the following classes:
        UserFieldDecls
        - "office:chart"
        - "office:drawing"
        - "office:presentation"
        - "office:spreadsheet"
        - "office:text"
        - "style:footer"
        - "style:footer-first"
        - "style:footer-left"
        - "style:header"
        - "style:header-first"
        - "style:header-left"
    """

    def get_user_field_decls(self) -> UserFieldDecls:
        """Returns the container for user field declarations.

        If the container is not found, it is created within the document body.

        Returns:
            UserFieldDecls: The UserFieldDecls instance (container for
                user field declarations).

        Raises:
            ValueError: If the document body is empty and a new container
                cannot be inserted.
        """
        user_field_decls = self.get_element("//text:user-field-decls")
        if user_field_decls is None:
            body = self.document_body
            if not body:
                raise ValueError("Empty document.body")
            body.insert(Element.from_tag("text:user-field-decls"), FIRST_CHILD)
            user_field_decls = body.get_element("//text:user-field-decls")

        return cast(UserFieldDecls, user_field_decls)


class UserFieldDecls(UserFieldDeclMixin):
    """A container for user field declarations, "text:user-field-decls".

    This element groups all the 'text:user-field-decl' elements in the
    document's meta information.
    """

    _tag = "text:user-field-decls"


class UserFieldDecl(ElementTyped):
    """A declaration of a user field, "text:user-field-decl".

    This element, typically found within 'text:user-field-decls', defines a
    user-defined field, its name, type, and current value.

    Attributes:
        name (str): The unique name of the user field.
    """

    _tag = "text:user-field-decl"
    _properties = (PropDef("name", "text:name"),)

    def __init__(
        self,
        name: str | None = None,
        value: Any = None,
        value_type: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the UserFieldDecl element.

        Args:
            name: The name of the user field.
            value: The initial value of the field.
            value_type: The ODF value type (e.g., 'string',
                'float'). If not provided, it is inferred from the `value`.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if name:
                self.name = name
            self.set_value_and_type(value=value, value_type=value_type)

    def set_value(self, value: Any) -> None:
        """Sets the value of the user field declaration.

        This method updates the value and value type of the declaration,
        preserving its name.

        Args:
            value: The new value for the field.
        """
        name = self.get_attribute("text:name")
        self.clear()
        self.set_value_and_type(value=value)
        self.set_attribute("text:name", name)


UserFieldDecl._define_attribut_property()


register_element_class(UserFieldDecls)
register_element_class(UserFieldDecl)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Mixin class NRMixin for classes using Named Ranges."""

from __future__ import annotations

from typing import Union, cast

from .element import Element, PropDef, PropDefBool, register_element_class
from .named_range import NamedRange


class TableNamedExpressions(Element):
    """Container of assignments of names to expressions, tag "table:named-expressions".

    (Mostly internal use).
    The following expressions may have names:
      - cell ranges,
      - Other expressions.

    If the "table:named-expressions" element is used with a "table:table" element,
    the scope of the named expressions are local to that table element.
    """

    _tag: str = "table:named-expressions"
    _properties: tuple[PropDef | PropDefBool, ...] = ()


class NRMixin(Element):
    """Mixin for Named Range access.

    Used by the following classes: Chart, Drawing, Presentation, Spreadsheet, Text.
    """

    def get_named_ranges(self) -> list[NamedRange]:
        """Retrieve all named ranges (`NamedRange` objects) global to the document.

        Returns:
            list[NamedRange]: A list of `NamedRange` instances found in the document.
        """
        return cast(
            list[NamedRange],
            self.get_elements("descendant::table:named-expressions/table:named-range"),
        )

    def get_named_range(self, name: str) -> NamedRange | None:
        """Retrieve a specific named range global to the document by its name.

        Args:
            name: The name of the named range to retrieve.

        Returns:
            NamedRange | None: The `NamedRange` instance if found, otherwise `None`.
        """
        named_range = cast(
            list[NamedRange],
            self.get_elements(
                f'descendant::table:named-expressions/table:named-range[@table:name="{name}"][1]'
            ),
        )
        if named_range:
            return named_range[0]
        else:
            return None

    def append_named_range(self, named_range: NamedRange) -> None:
        """Append a `NamedRange` object to the document.

        If a named range with the same name already exists, it will be replaced.

        Args:
            named_range: The `NamedRange` object to append.
        """
        named_expressions = cast(
            Union[None, TableNamedExpressions],
            self.get_element("table:named-expressions"),
        )
        if not named_expressions:
            named_expressions = cast(
                TableNamedExpressions, Element.from_tag("table:named-expressions")
            )
            self._Element__append(named_expressions)  # type: ignore[attr-defined]
        # exists ?
        current = cast(
            Union[None, NamedRange],
            named_expressions.get_element(
                f'table:named-range[@table:name="{named_range.name}"][1]'
            ),
        )
        if current:
            named_expressions.delete(current)
        named_expressions._Element__append(named_range)  # type: ignore[attr-defined]

    def set_named_range(
        self,
        name: str,
        crange: str | tuple | list,
        table_name: str,
        usage: str | None = None,
    ) -> None:
        """Create or update a named range in the document.

        A `NamedRange` element is created with the given parameters and inserted
        into the document. If a named range with the same name already exists,
        it will be replaced.

        Args:
            name: The name of the named range.
            crange: The cell or cell range coordinate
                (e.g., "Sheet1.A1", "Sheet1.A1:B2", or a tuple/list of integers
                representing coordinates).
            table_name: The name of the table to which the named range refers.
            usage: Optional usage type (e.g., "print-range", "filter",
                "repeat-column", "repeat-row").

        Raises:
            ValueError: If `name` or `table_name` is empty.
        """
        name = name.strip()
        if not name:
            raise ValueError("Name required")
        table_name = table_name.strip()
        if not table_name:
            raise ValueError("Table name required")
        named_range = NamedRange(name, crange, table_name, usage)
        self.append_named_range(named_range)

    def delete_named_range(self, name: str) -> None:
        """Delete a named range from the document by its name.

        Args:
            name: The name of the named range to delete.
        """
        named_range = self.get_named_range(name)
        if not named_range:
            return
        named_range.delete()
        named_expressions = cast(
            TableNamedExpressions, self.get_element("table:named-expressions")
        )
        if named_expressions.is_empty():
            self.delete(named_expressions)


register_element_class(TableNamedExpressions)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

import base64
import io

from odfdo.document import Blob


def test_blob_null():
    blob = Blob()
    assert blob.content == b""
    assert blob.name == ""
    assert blob.mime_type == ""


def test_blob_png(samples):
    png = samples("image.png")
    blob = Blob.from_path(png)
    assert blob.name == "ceddccf10506d07cc0990639e79f8c72.png"
    assert blob.mime_type == "image/png"


def test_blob_png_content(samples):
    png = samples("image.png")
    expected = png.read_bytes()
    blob = Blob.from_path(png)
    assert blob.content == expected


def test_blob_png_str(samples):
    png = str(samples("image.png"))
    blob = Blob.from_path(png)
    assert blob.name == "ceddccf10506d07cc0990639e79f8c72.png"
    assert blob.mime_type == "image/png"


def test_blob_png_content_str(samples):
    png = samples("image.png")
    expected = png.read_bytes()
    blob = Blob.from_path(str(png))
    assert blob.content == expected


def test_blob_jpg(samples):
    png = samples("image2.jpg")
    blob = Blob.from_path(png)
    assert blob.name == "b011604234764865b75f7dbd903992de.jpg"
    assert blob.mime_type == "image/jpeg"


def test_blob_jpg_content(samples):
    png = samples("image2.jpg")
    expected = png.read_bytes()
    blob = Blob.from_path(png)
    assert blob.content == expected


def test_blob_io(samples):
    png = samples("image.png")
    with io.BytesIO() as bytes_content:
        bytes_content.write(png.read_bytes())
        bytes_content.seek(0)
        blob = Blob.from_io(bytes_content)
    assert blob.name == "ceddccf10506d07cc0990639e79f8c72"
    assert blob.mime_type == "application/octet-stream"


def test_blob_io_content(samples):
    png = samples("image.png")
    expected = png.read_bytes()
    with io.BytesIO() as bytes_content:
        bytes_content.write(png.read_bytes())
        bytes_content.seek(0)
        blob = Blob.from_io(bytes_content)
    assert blob.content == expected


def test_blob_b64(samples):
    png = samples("image.png")
    content = png.read_bytes()
    b64string = base64.standard_b64encode(content)
    blob = Blob.from_base64(b64string, "image/png")
    assert blob.name == "ceddccf10506d07cc0990639e79f8c72"
    assert blob.mime_type == "image/png"


def test_blob_b64_content(samples):
    png = samples("image.png")
    content = png.read_bytes()
    b64string = base64.standard_b64encode(content)
    blob = Blob.from_base64(b64string, "image/png")
    assert blob.content == content



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>

from collections.abc import Iterable

import pytest

from odfdo.document import Document
from odfdo.element import Element


@pytest.fixture
def body(samples) -> Iterable[Element]:
    document = Document(samples("example.odt"))
    yield document.body


def test_first(body):
    last_paragraph = body.get_paragraph(position=0)
    expected = "This is the first paragraph."
    assert last_paragraph.inner_text == expected


def test_next_to_last(body):
    last_paragraph = body.get_paragraph(position=-2)
    expected = "This is an annotation."
    assert last_paragraph.inner_text == expected


def test_last(body):
    last_paragraph = body.get_paragraph(position=-1)
    expected = "With diacritical signs: éè"
    assert last_paragraph.inner_text == expected



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Mixin class for classes containing Lists, "text:list"."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union, cast

from .element import Element

if TYPE_CHECKING:
    from .list import List


class ListMixin(Element):
    """Mixin class for classes containing Lists.

    Used by the following classes:
    *"draw:caption", *"draw:circle", "draw:connector", *"draw:custom-shape",
    "draw:ellipse", "draw:image", "draw:line", *"draw:measure", *"draw:path",
    *"draw:polygon", *"draw:polyline", "draw:rect", *"draw:regular-polygon",
    "draw:text-box", "office:annotation", "office:text", "style:footer",
    "style:footer-first", "style:footer-left", "style:header",
    "style:header-first", "style:header-left", "table:covered-table-cell",
    "table:table-cell", "text:deletion", "text:index-body",
    "text:index-title", "text:list-header", "text:list-item",
    "text:note-body", "text:section".

    *: not implemented.
    """

    def get_lists(
        self,
        style: str | None = None,
        content: str | None = None,
    ) -> list[List]:
        """Returns all lists that match the specified criteria.

        Args:
            style: The name of the style to filter lists by.
            content: A regex pattern to match against the list's content.

        Returns:
            list[List]: A list of `List` instances matching the criteria.
        """
        return cast(
            "list[List]",
            self._filtered_elements(
                "descendant::text:list", text_style=style, content=content
            ),
        )

    @property
    def lists(self) -> list[List]:
        """Returns all lists as a list.

        Returns:
            list[List]: A list of all List instances that are descendants of
                this element.
        """
        return cast("list[List]", self.get_elements("descendant::text:list"))

    def get_list(
        self,
        position: int = 0,
        content: str | None = None,
    ) -> List | None:
        """Returns a single list that matches the specified criteria.

        Args:
            position: The 0-based index of the matching list to return.
            content: A regex pattern to match against the list's content.

        Returns:
            List | None: A List instance, or None if no list matches the
                criteria.
        """
        return cast(
            "Union[None, List]",
            self._filtered_element("descendant::text:list", position, content=content),
        )



#!/usr/bin/env python
"""Minimal example of how to add a styled paragraph to a document."""

import os

from odfdo import Document, Paragraph

_DOC_SEQUENCE = 335


def generate_document() -> Document:
    """Return a document with some styled paragraph"""
    document = Document("text")
    body = document.body
    body.clear()

    # Assuming we have a style of name "highlight" :
    body.append(Paragraph("Highlighting the word", style="highlight"))

    return document


def main() -> None:
    document = generate_document()
    test_unit(document)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    paragraphs = document.body.paragraphs
    assert len(paragraphs) == 1
    assert paragraphs[0].style == "highlight"


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Authors (odfdo project): jerome.dumonteil@gmail.com
from __future__ import annotations

import io
import subprocess
import sys
from pathlib import Path

import pytest

from odfdo.document import Document
from odfdo.scripts import table_shrink
from odfdo.scripts.table_shrink import main as main_script
from odfdo.scripts.table_shrink import main_shrink, parse_cli_args

SCRIPT = Path(table_shrink.__file__)


def run_params(params: list):
    command = [sys.executable, SCRIPT] + params
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_shrink_no_param():
    params = []
    out, _err, exitcode = run_params(params)
    assert exitcode == 1
    assert b"usage: odfdo-table-shrink" in out


# direct access to internal function


def test_shrink_2_no_param_on_main_function(monkeypatch):
    with pytest.raises(SystemExit) as result:
        monkeypatch.setattr(sys, "argv", [])
        main_script()
        assert result.value.code >= 1


def test_shrink_2_no_param():
    with pytest.raises(SystemExit) as result:
        params = parse_cli_args([])
        main_shrink(params)
        assert result.value.code >= 1


def test_shrink_2_version(capsys):
    with pytest.raises(SystemExit) as result:
        parse_cli_args(["--version"])
        assert result.value.code == 0
    captured = capsys.readouterr()

    assert "odfdo-table-shrink v3" in captured.out


def test_shrink_2_help(capsys):
    with pytest.raises(SystemExit) as result:
        parse_cli_args(["--help"])
        assert result.value.code == 0
    captured = capsys.readouterr()

    assert "Optimize the width" in captured.out


def test_shrink_2_no_file():
    params = parse_cli_args(["-i", "none_file1", "-o", "none_file2"])
    with pytest.raises(SystemExit) as result:
        main_shrink(params)
        assert result.value.code >= 1


def test_shrink_2_base_on_main_function(capsysbinary, monkeypatch, samples):
    source = str(samples("test_col_cell_blue.ods"))
    monkeypatch.setattr(sys, "argv", ["odfdo-table-shrink", "-i", source])

    main_script()
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    table = document.body.get_table(0)
    assert table.size == (3, 3)


def test_shrink_2_trim_1(capsysbinary, samples):
    source = str(samples("test_col_cell_blue.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    table = document.body.get_table(0)
    assert table.size == (3, 3)


def test_shrink_2_trim_2(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    table = document.body.get_table(0)
    assert table.size == (7, 16)


def test_shrink_2_trim_3(capsysbinary, samples):
    source = str(samples("simple_table_named_range.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    table = document.body.get_table(0)
    assert table.size == (7, 4)
    table = document.body.get_table(1)
    assert table.size == (1, 1)
    table = document.body.get_table(2)
    assert table.size == (2, 2)


def test_shrink_2_trim_4(capsysbinary, samples):
    source = str(samples("simple_table.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    table = document.body.get_table(0)
    assert table.size == (7, 4)
    table = document.body.get_table(1)
    assert table.size == (1, 1)
    table = document.body.get_table(2)
    assert table.size == (2, 2)


def test_shrink_2_trim_5(capsysbinary, samples):
    source = str(samples("styled_table.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    table = document.body.get_table(0)
    assert table.size == (5, 10)
    table = document.body.get_table(1)
    assert table.size == (1, 1)
    table = document.body.get_table(2)
    assert table.size == (1, 1)


def test_shrink_2_color_1_1(capsysbinary, samples):
    source = str(samples("test_col_cell_blue.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "a1") == "#2a6099"
    assert document.get_cell_background_color(0, "b1") == "#2a6099"
    assert document.get_cell_background_color(0, "c1") == "#2a6099"
    assert document.get_cell_background_color(0, "d1") == "#ffffff"
    assert document.get_cell_background_color(0, (1000, 0)) == "#ffffff"


def test_shrink_2_color_1_2(capsysbinary, samples):
    source = str(samples("test_col_cell_blue.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "a2") == "#2a6099"
    assert document.get_cell_background_color(0, "b2") == "#ffff00"
    assert document.get_cell_background_color(0, "c2") == "#2a6099"
    assert document.get_cell_background_color(0, "d2") == "#ffffff"
    assert document.get_cell_background_color(0, (1000, 1)) == "#ffffff"


def test_shrink_2_color_1_3(capsysbinary, samples):
    source = str(samples("test_col_cell_blue.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "a3") == "#2a6099"
    assert document.get_cell_background_color(0, "b3") == "#2a6099"
    assert document.get_cell_background_color(0, "c3") == "#2a6099"
    assert document.get_cell_background_color(0, "d3") == "#ffffff"
    assert document.get_cell_background_color(0, (1000, 2)) == "#ffffff"


def test_shrink_2_color_2_1(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "a1") == "#ffffff"
    assert document.get_cell_background_color(0, "c1") == "#ffff00"
    assert document.get_cell_background_color(0, "d1") == "#ffffff"
    assert document.get_cell_background_color(0, "f1") == "#e6e905"
    assert document.get_cell_background_color(0, "g1") == "#ffffff"
    assert document.get_cell_background_color(0, (1000, 0)) == "#ffffff"


def test_shrink_2_color_2_2(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "a2") == "#ffffff"
    assert document.get_cell_background_color(0, "c2") == "#ff3838"
    assert document.get_cell_background_color(0, "f2") == "#e6e905"
    assert document.get_cell_background_color(0, "g2") == "#ffffff"
    assert document.get_cell_background_color(0, (1000, 1)) == "#ffffff"


def test_shrink_2_color_2_5(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "f5") == "#e6e905"


def test_shrink_2_color_2_6(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "a6") == "#3465a4"
    assert document.get_cell_background_color(0, "c6") == "#3465a4"
    assert document.get_cell_background_color(0, "f6") == "#e6e905"
    assert document.get_cell_background_color(0, "g6") == "#3465a4"
    assert document.get_cell_background_color(0, "h6") == "#ffffff"
    assert document.get_cell_background_color(0, (1000, 5)) == "#ffffff"


def test_shrink_2_color_2_7(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "f7") == "#e6e905"


def test_shrink_2_color_2_8(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "a8") == "#b4c7dc"
    assert document.get_cell_background_color(0, "b8") == "#b4c7dc"
    assert document.get_cell_background_color(0, "c8") == "#ffff6d"
    assert document.get_cell_background_color(0, "d8") == "#ff3838"
    assert document.get_cell_background_color(0, "e8") == "#b4c7dc"
    assert document.get_cell_background_color(0, "f8") == "#e6e905"
    assert document.get_cell_background_color(0, "g8") == "#b4c7dc"
    assert document.get_cell_background_color(0, "h8") == "#ffffff"
    assert document.get_cell_background_color(0, (1000, 5)) == "#ffffff"


def test_shrink_2_color_2_9(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "f9") == "#e6e905"


def test_shrink_2_color_2_10(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "a10") == "#ed4c05"
    assert document.get_cell_background_color(0, "b10") == "#ed4c05"
    assert document.get_cell_background_color(0, "c10") == "#ed4c05"
    assert document.get_cell_background_color(0, "d10") == "#ed4c05"
    assert document.get_cell_background_color(0, "e10") == "#ed4c05"
    assert document.get_cell_background_color(0, "f10") == "#e6e905"
    assert document.get_cell_background_color(0, "g10") == "#ed4c05"
    assert document.get_cell_background_color(0, "h10") == "#ffffff"
    assert document.get_cell_background_color(0, (1000, 9)) == "#ffffff"


def test_shrink_2_color_2_11(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "f11") == "#e6e905"


def test_shrink_2_color_2_12(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "a12") == "#168253"
    assert document.get_cell_background_color(0, "b12") == "#168253"
    assert document.get_cell_background_color(0, "c12") == "#168253"
    assert document.get_cell_background_color(0, "d12") == "#168253"
    assert document.get_cell_background_color(0, "e12") == "#168253"
    assert document.get_cell_background_color(0, "f12") == "#e6e905"
    assert document.get_cell_background_color(0, "g12") == "#168253"
    assert document.get_cell_background_color(0, "h12") == "#ffffff"
    assert document.get_cell_background_color(0, (1000, 11)) == "#ffffff"


def test_shrink_2_color_2_13(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "f13") == "#e6e905"


def test_shrink_2_color_2_14(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "f14") == "#fff5ce"


def test_shrink_2_color_2_15(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "a15") == "#55215b"
    assert document.get_cell_background_color(0, "b15") == "#55215b"
    assert document.get_cell_background_color(0, "c15") == "#55215b"
    assert document.get_cell_background_color(0, "d15") == "#55215b"
    assert document.get_cell_background_color(0, "e15") == "#55215b"
    assert document.get_cell_background_color(0, "f15") == "#55215b"
    assert document.get_cell_background_color(0, "g15") == "#55215b"
    assert document.get_cell_background_color(0, "h15") == "#ffffff"
    assert document.get_cell_background_color(0, (1000, 14)) == "#ffffff"


def test_shrink_2_color_2_16(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "f16") == "#e6e905"


def test_shrink_2_color_2_17(capsysbinary, samples):
    source = str(samples("test_col_cell.ods"))
    params = parse_cli_args(["-i", source])

    main_shrink(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    assert document.get_cell_background_color(0, "f17") == "#ffffff"


def test_shrink_2_bad_type(samples):
    source = str(samples("base_text.odt"))
    params = parse_cli_args(["-i", source])

    with pytest.raises(SystemExit) as result:
        main_shrink(params)
        assert result.value.code >= 1



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
"""Classes related to user-defined fields in ODF documents.

This module provides classes for managing user field declarations
content within the document, "text:user-defined", text:user-field-get",
"text:user-field-input".
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union, cast

from .element import Element, PropDef, register_element_class
from .element_typed import ElementTyped

# for compatibility for version <= 3.18.2
from .user_field_declaration import (  # noqa: F401
    UserFieldDecl,
    UserFieldDecls,
)

if TYPE_CHECKING:
    from datetime import datetime, timedelta
    from decimal import Decimal

    from .document import Document


class UserDefinedMixin(Element):
    """Mixin class for classes containing user defined fields.

    Used by the following classes:
        - "text:a"
        - "text:h"
        - "text:meta"
        - "text:meta-field"
        - "text:p"
        - "text:ruby-base"
        - "text:span"

    and "office:text", "office:spreadsheet","office:presentation" for
    compatibility with previous versions.
    """

    def get_user_defined_list(self) -> list[UserDefined]:
        """Returns all user-defined field declarations as a list.

        Returns:
            list[UserDefined]: A list of all UserDefined instances that are descendants of this element.
        """
        return cast(
            list[UserDefined],
            self._filtered_elements(
                "descendant::text:user-defined",
            ),
        )

    @property
    def user_defined_list(self) -> list[UserDefined]:
        """Returns all user-defined field declarations as a list.

        Returns:
            list[UserDefined]: A list of all UserDefined instances that are descendants of this element.
        """
        return self.get_user_defined_list()

    def get_user_defined(self, name: str, position: int = 0) -> UserDefined | None:
        """Returns a single user-defined field declaration that matches the specified criteria.

        Args:
            name: The name of the user-defined field to retrieve.
            position: The 0-based index of the matching user-defined field to return.

        Returns:
            UserDefined | None: A UserDefined instance, or None if no declaration matches the criteria.
        """
        return cast(
            Union[None, UserDefined],
            self._filtered_element(
                "descendant::text:user-defined", position, text_name=name
            ),
        )

    def get_user_defined_value(
        self, name: str, value_type: str | None = None
    ) -> bool | str | int | float | Decimal | datetime | timedelta | None:
        """Returns the value of the specified user-defined field.

        Args:
            name: The name of the user-defined field to retrieve its value.
            value_type: The expected type of the user-defined field's value.
                Can be 'boolean', 'date', 'float', 'string', 'time', or None
                for automatic type detection.

        Returns:
            bool | str | int | float | Decimal | datetime | timedelta | None:
                The value of the user-defined field, cast to the most appropriate
                Python type, or None if the user-defined field is not found.
        """
        user_defined = self.get_user_defined(name)
        if user_defined is None:
            return None
        return user_defined.get_value(value_type)  # type: ignore[return-value]


class UserFieldGet(ElementTyped):
    """A getter for a user field value, "text:user-field-get".

    This element displays the current value of a user-defined field at a
    specific position in the document content.

    Attributes:
        name (str): The name of the user field to display.
        style (str, optional): The data style to apply for formatting the
            displayed value.
    """

    _tag = "text:user-field-get"
    _properties = (
        PropDef("name", "text:name"),
        PropDef("style", "style:data-style-name"),
    )

    def __init__(
        self,
        name: str | None = None,
        value: Any = None,
        value_type: str | None = None,
        text: str | None = None,
        style: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the UserFieldGet element.

        Args:
            name: The name of the user field to get.
            value: An initial value to display.
            value_type: The ODF value type.
            text: The textual representation to display. If
                not provided, it's generated from `value`.
            style: The data style name for formatting.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if name:
                self.name = name
            text = self.set_value_and_type(
                value=value, value_type=value_type, text=text
            )
            self.text = text

            if style:
                self.style = style


UserFieldGet._define_attribut_property()


class UserFieldInput(UserFieldGet):
    """An input field for a user-defined field, "text:user-field-input".

    This element allows the user to interactively modify the value of a
    user-defined field within the document.
    """

    _tag = "text:user-field-input"


UserFieldInput._define_attribut_property()


class UserDefined(ElementTyped):
    """A user-defined field instance, "text:user-defined".

    This element is similar to 'text:user-field-get' but is often used in
    contexts like headers or footers. It can be initialized directly with
    a value or can pull its value from a corresponding user-defined metadata
    field in the document's meta section.

    Attributes:
        name (str): The name of the user field.
        style (str, optional): The data style to apply for formatting.
    """

    _tag = "text:user-defined"
    _properties = (
        PropDef("name", "text:name"),
        PropDef("style", "style:data-style-name"),
    )

    def __init__(
        self,
        name: str = "",
        value: Any = None,
        value_type: str | None = None,
        text: str | None = None,
        style: str | None = None,
        from_document: Document | None = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the UserDefined element.

        If a document is provided via `from_document`, the element will be
        populated with the value of the meta user-defined field of the same
        name from that document.

        Args:
            name: The name of the user-defined field.
            value: The value of the field.
            value_type: The ODF value type (e.g., 'string').
            text: The textual representation of the value.
            style: The data style name for formatting.
            from_document: A document from which to load
                the field's value from the meta section.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if name:
                self.name = name
            if style:
                self.style = style
            if from_document is not None:
                meta_infos = from_document.meta
                content = meta_infos.get_user_defined_metadata_of_name(name)
                if content is not None:
                    value = content.get("value", None)
                    value_type = content.get("value_type", None)
                    text = content.get("text", None)
            text = self.set_value_and_type(
                value=value, value_type=value_type, text=text
            )
            self.text = text


UserDefined._define_attribut_property()

register_element_class(UserFieldGet)
register_element_class(UserFieldInput)
register_element_class(UserDefined)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""MetaAutoReload class for "meta:auto-reload" tag."""

from __future__ import annotations

from datetime import timedelta
from typing import Any

from .datatype import Duration
from .element import Element, PropDef, PropDefBool, register_element_class


class MetaAutoReload(Element):
    """Container for auto-reload properties, "meta:auto-reload"."""

    _tag = "meta:auto-reload"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("delay", "meta:delay"),
        PropDef("actuate", "xlink:actuate"),
        PropDef("href", "xlink:href"),
        PropDef("show", "xlink:show"),
        PropDef("type", "xlink:type"),
    )

    def __init__(
        self,
        delay: timedelta | None = None,
        href: str = "",
        **kwargs: Any,
    ) -> None:
        """Initialize a MetaAutoReload element.

        The `meta:auto-reload` element specifies whether a document is
        reloaded or replaced by another document after a specified period
        of time has elapsed.

        Args:
            delay: The time delay after which the document
                should auto-reload.
            href: The URL or path to the document to reload or replace with.
            **kwargs: Additional keyword arguments for the parent `Element` class.
        """
        super().__init__(**kwargs)

        self.actuate = "onLoad"
        self.show = "replace"
        self.type = "simple"
        if self._do_init:
            self._set_delay(delay)
            self.href = href

    def __repr__(self) -> str:
        if self.delay:
            delay = str(Duration.decode(self.delay))
        else:
            delay = ""
        return (
            f"<{self.__class__.__name__} tag={self.tag} href={self.href} delay={delay}>"
        )

    def __str__(self) -> str:
        return f"({self.href})"

    def _set_delay(self, delay: timedelta | None) -> None:
        """Set the `meta:delay` attribute from a `timedelta` object.

        Converts the `timedelta` to an ODF duration string. If `delay` is
        None, it defaults to `timedelta(0)`.

        Args:
            delay: The delay duration.
        """
        if delay is None:
            delay = timedelta(0)
        self.delay = Duration.encode(delay)

    def as_dict(self) -> dict[str, Any]:
        """Return the attributes of the auto-reload element as a Python dictionary.

        Returns:
            dict[str, Any]: A dictionary containing the meta auto-reload
                attributes, with keys like "meta:delay", "xlink:href", etc.
        """
        result: dict[str, Any] = {}
        if self.delay:
            result["meta:delay"] = Duration.decode(self.delay)
        else:
            result["meta:delay"] = timedelta(0)
        result.update(
            {
                "xlink:actuate": self.actuate or "onLoad",
                "xlink:href": self.href or "",
                "xlink:show": self.show or "replace",
                "xlink:type": self.type or "simple",
            }
        )
        return result

    def from_dict(self, data: dict[str, Any]) -> None:
        """Set the attributes of the auto-reload element from a Python dictionary.

        Args:
            data: A dictionary containing the meta auto-reload
                attributes (e.g., "meta:delay", "xlink:href").
        """
        self._set_delay(data.get("meta:delay"))
        self.actuate = data.get("xlink:actuate", "onLoad")
        self.href = data.get("xlink:href", "")
        self.show = data.get("xlink:show", "replace")
        self.type = data.get("xlink:type", "simple")


MetaAutoReload._define_attribut_property()

register_element_class(MetaAutoReload)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.form_controls import FormTextarea
from odfdo.paragraph import Paragraph


@pytest.fixture
def form_textarea() -> Iterable[FormTextarea]:
    yield FormTextarea(
        name="some name",
        control_implementation="some implem",
        value="initial",
        current_value="final",
        convert_empty_to_null=False,
        data_field="",
        linked_cell="",
        readonly=False,
        disabled=False,
        printable=True,
        tab_index=4,
        tab_stop=False,
        xml_id="control1",
    )


@pytest.fixture
def form_textarea2() -> Iterable[FormTextarea]:
    form = FormTextarea(
        name="name2",
        control_implementation="some implem",
        value="initial2",
        current_value="final2",
        convert_empty_to_null=False,
        data_field="",
        linked_cell="",
        readonly=False,
        disabled=False,
        printable=True,
        tab_index=4,
        tab_stop=False,
        xml_id="control2",
    )
    p1 = Paragraph("First paragraph")
    p2 = Paragraph("Second paragraph")
    form.append(p1)
    form.append(p2)
    yield form


def test_form_textarea_class():
    form = FormTextarea()
    assert isinstance(form, FormTextarea)


def test_form_textarea_minimal_tag():
    form = Element.from_tag("<form:textarea/>")
    assert isinstance(form, FormTextarea)


def test_form_textarea_serialize():
    form = FormTextarea()
    assert form.serialize() == "<form:textarea/>"


def test_form_textarea_repr():
    form = FormTextarea()
    assert repr(form) == "<FormTextarea name=None xml_id=None>"


def test_form_textarea_name(form_textarea):
    assert form_textarea.name == "some name"


def test_form_textarea_control_implementation(form_textarea):
    assert form_textarea.control_implementation == "some implem"


def test_form_textarea_disabled(form_textarea):
    assert form_textarea.disabled is False


def test_form_textarea_printable(form_textarea):
    assert form_textarea.printable is True


def test_form_textarea_tab_index(form_textarea):
    assert form_textarea.tab_index == 4


def test_form_textarea_tab_stop(form_textarea):
    assert form_textarea.tab_stop is False


def test_form_textarea_xml_id(form_textarea):
    assert form_textarea.xml_id == "control1"


def test_form_textarea_form_id(form_textarea):
    assert form_textarea.form_id == "control1"


def test_form_textarea_xforms_bind():
    form = FormTextarea(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        xforms_bind="bind id",
    )
    assert form.xforms_bind == "bind id"


def test_form_textarea_form_id_2():
    form = FormTextarea(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
    )
    assert form.form_id == "control1"


def test_form_textarea_form_value(form_textarea):
    assert form_textarea.value == "initial"


def test_form_textarea_form_current_value(form_textarea):
    assert form_textarea.current_value == "final"


def test_form_text_area_form_as_dict(form_textarea):
    form_textarea.current_value = None
    expected = {
        "tag": "form:textarea",
        "name": "some name",
        "xml_id": "control1",
        "value": "initial",
        "current_value": None,
        "str": "",
    }
    assert form_textarea.as_dict() == expected


def test_form_textarea_form_current_value_2(form_textarea):
    form_textarea.current_value = None
    assert form_textarea.current_value is None


def test_form_textarea_convert_empty_to_null(form_textarea):
    assert form_textarea.convert_empty_to_null is False


def test_form_textarea_data_field(form_textarea):
    assert form_textarea.data_field == ""


def test_form_textarea_linked_cell(form_textarea):
    assert form_textarea.linked_cell == ""


def test_form_textarea_readonly(form_textarea):
    assert form_textarea.readonly is False


def test_form_textarea_value2(form_textarea2):
    assert form_textarea2.value == "initial2"


def test_form_textarea_current_value2(form_textarea2):
    assert form_textarea2.current_value == "final2"


def test_form_text_area_form_as_dict2(form_textarea):
    form_textarea.current_value = None
    expected = {
        "tag": "form:textarea",
        "name": "some name",
        "xml_id": "control1",
        "value": "initial",
        "current_value": None,
        "str": "",
    }
    assert form_textarea.as_dict() == expected


def test_form_textarea_str(form_textarea):
    assert str(form_textarea) == ""


def test_form_textarea_str2(form_textarea2):
    assert str(form_textarea2) == "First paragraph\nSecond paragraph\n"


def test_form_text_area_form_as_dict3(form_textarea2):
    expected = {
        "tag": "form:textarea",
        "name": "name2",
        "xml_id": "control2",
        "value": "initial2",
        "current_value": "final2",
        "str": "First paragraph\nSecond paragraph\n",
    }
    assert form_textarea2.as_dict() == expected



# Copyright 2018-2026 Jérôme Dumonteil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.document import Document
from odfdo.table import Table


@pytest.fixture
def body(samples) -> Iterable[Element]:
    document = Document(samples("simple_table.ods"))
    yield document.body


@pytest.fixture
def table(samples) -> Iterable[Table]:
    # simpletable :
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1   2	3	4	5	6	7
    document = Document(samples("simple_table.ods"))
    yield document.body.get_table(name="Example1")


def test_table_empty():
    table = Table("empty")
    table.transpose()
    assert table.get_values() == []


def test_table_empty_2():
    table = Table("empty", 3, 2)
    table.transpose()
    assert table.get_values() == [
        [None, None],
        [None, None],
        [None, None],
    ]


def test_table_empty_3():
    table = Table("empty", 3, 2)
    table.transpose((0, 0, 1, 1))
    assert table.get_values() == [
        [None, None, None],
        [None, None, None],
    ]


# simpletable :
#   1	1	1	2	3	3	3
#   1	1	1	2	3	3	3
#   1	1	1	2	3	3	3
#   1   2	3	4	5	6	7


def test_table_transpose(table):
    table.transpose()
    assert table.get_values() == [
        [1, 1, 1, 1],
        [1, 1, 1, 2],
        [1, 1, 1, 3],
        [2, 2, 2, 4],
        [3, 3, 3, 5],
        [3, 3, 3, 6],
        [3, 3, 3, 7],
    ]


def test_table_transpose_1(table):
    coord = (None, None, None, None)
    table.transpose(coord)
    result = list(table.iter_values(complete=True))
    assert result == [
        [1, 1, 1, 1, None, None, None],
        [1, 1, 1, 2, None, None, None],
        [1, 1, 1, 3, None, None, None],
        [2, 2, 2, 4, None, None, None],
        [3, 3, 3, 5, None, None, None],
        [3, 3, 3, 6, None, None, None],
        [3, 3, 3, 7, None, None, None],
    ]


def test_table_transpose_2(table):
    table.transpose("A1:G1")
    assert table.get_values() == [
        [1, None, None, None, None, None, None],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [2, 2, 3, 4, 5, 6, 7],
        [3, None, None, None, None, None, None],
        [3, None, None, None, None, None, None],
        [3, None, None, None, None, None, None],
    ]


def test_table_transpose_3(table):
    table.delete_row(3)
    table.delete_row(2)
    table.delete_row(1)
    table.transpose()
    assert table.get_values() == [[1], [1], [1], [2], [3], [3], [3]]


def test_table_transpose_3c(table):
    table.delete_column(6)
    table.delete_column(5)
    table.delete_column(4)
    table.delete_column(3)
    table.delete_column(2)
    table.delete_column(1)
    table.transpose()
    assert table.get_values() == [[1, 1, 1, 1]]


def test_table_transpose_4(table):
    table.transpose("F2:F4")
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3, None],
        [1, 1, 1, 2, 3, 3, 3, 6],
        [1, 1, 1, 2, 3, None, 3, None],
        [1, 2, 3, 4, 5, None, 7, None],
    ]



#!/usr/bin/env python
"""Create a spreadsheet with two tables, using named ranges to fill cells."""

import os
from pathlib import Path

from odfdo import Document, Table

_DOC_SEQUENCE = 470
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "named_range"
TARGET = "spreadsheet.ods"


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


def make_some_table_content() -> Table:
    table = Table("First Table")
    # populate the table :
    for index in range(10):
        table.set_value((1, index), (index + 1) ** 2)
    table.set_value("A11", "Total:")
    return table


def generate_document() -> Document:
    document = Document("spreadsheet")
    body = document.body
    body.clear()

    table = make_some_table_content()
    body.append(table)

    # lets define a named range for the 10 values :
    range_squares = "B1:B10"
    name = "squares_values"
    table_name = table.name
    table.set_named_range(name, range_squares, table_name)

    # we can define a single cell range, using notation "B11" or (1, 10) :
    table.set_named_range("total", (1, 10), table_name)

    # get named range values :
    values = table.get_named_range("squares_values").get_values(flat=True)

    # set named range value :
    result = sum(values)
    table.get_named_range("total").set_value(result)

    # lets use the named ranges from a second table :
    table2 = Table("Second Table")
    body.append(table2)

    named_range1 = table2.get_named_range("total")
    table2.set_value("A1", "name:")
    table2.set_value("B1", named_range1.name)
    table2.set_value("A2", "range:")
    table2.set_value("B2", str(named_range1.crange))
    table2.set_value("A3", "from table:")
    table2.set_value("B3", named_range1.table_name)
    table2.set_value("A4", "content:")
    table2.set_value("B4", named_range1.get_value())

    named_range2 = table2.get_named_range("squares_values")
    table2.set_value("D1", "name:")
    table2.set_value("E1", named_range2.name)
    table2.set_value("D2", "range:")
    table2.set_value("E2", str(named_range2.crange))
    table2.set_value("D3", "from table:")
    table2.set_value("E3", named_range2.table_name)
    table2.set_value("D4", "content:")
    # using "E4:4" notaion is a little hack for the area starting at E4 on row 4
    table2.set_values(values=[named_range2.get_values(flat=True)], coord="E4:4")

    print("Content of the table 1:")
    print(table.name)
    print(table.to_csv())
    print("Content of the table 2:")
    print(table2.name)
    print(table2.to_csv())

    return document


def main() -> None:
    document = generate_document()
    test_unit(document)
    save_new(document, TARGET)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    table1 = document.body.get_table(0)
    assert table1.get_cell("B1").value == 1
    assert table1.get_cell("B10").value == 100
    assert table1.get_cell("B11").value == 385
    table2 = document.body.get_table(1)
    assert table2.get_cell("B4").value == 385
    assert table2.get_cell("N4").value == 100


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from __future__ import annotations

from collections.abc import Iterable
from typing import cast

import pytest

from odfdo.document import Document
from odfdo.row import Row
from odfdo.table import Table


@pytest.fixture
def table(samples) -> Iterable[Table]:
    # simpletable :
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1   2	3	4	5	6	7
    document = Document(samples("simple_table.ods"))
    yield cast(Table, document.body.get_table(name="Example1"))


@pytest.fixture
def styled_table(samples) -> Iterable[Table]:
    document = Document(samples("styled_table.ods"))
    yield cast(Table, document.body.get_table(name="Feuille1"))


@pytest.fixture
def unstriped_table_1(samples) -> Iterable[Table]:
    document = Document(samples("unstriped.ods"))
    yield cast(Table, document.body.get_table(name="Sheet1"))


@pytest.fixture
def unstriped_table_2(samples) -> Iterable[Table]:
    document = Document(samples("unstriped.ods"))
    yield cast(Table, document.body.get_table(name="Sheet2"))


def test_table_rstrip_unstriped_table_1(unstriped_table_1):
    unstriped_table_1.rstrip()
    assert unstriped_table_1.size[0] > 1000
    assert unstriped_table_1.size[1] > 1000


def test_table_rstrip_unstriped_table_1_gg(unstriped_table_1):
    unstriped_table_1.rstrip(aggressive=True)
    assert unstriped_table_1.size == (4, 4)


def test_table_rstrip_unstriped_table_2(unstriped_table_2):
    unstriped_table_2.rstrip()
    assert unstriped_table_2.size[0] == 6
    assert unstriped_table_2.size[1] == 5


def test_table_rstrip_styled_table(styled_table):
    styled_table.rstrip()
    assert styled_table.size == (5, 9)


def test_table_rstrip_1(table):
    table.rstrip()
    result = list(table.iter_values(complete=True))
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_rstrip_2(table):
    col = table.columns[6]
    col.repeated = 4
    table.set_column(6, col)
    result = list(table.iter_values(complete=True))
    assert result == [
        [1, 1, 1, 2, 3, 3, 3, None, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None, None],
        [1, 2, 3, 4, 5, 6, 7, None, None, None],
    ]


def test_table_rstrip_3(table):
    col = table.columns[6]
    col.repeated = 4
    table.set_column(6, col)
    table.rstrip()
    result = list(table.iter_values(complete=True))
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_rstrip_4(table):
    col = table.columns[6]
    col.repeated = 4
    table.set_column(6, col)
    row = Row()
    row.set_values([10, 20, 30, 40, 50, 60, 70, 80, 90])
    table.append(row)
    result = list(table.iter_values(complete=True))
    assert result == [
        [1, 1, 1, 2, 3, 3, 3, None, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None, None],
        [1, 2, 3, 4, 5, 6, 7, None, None, None],
        [10, 20, 30, 40, 50, 60, 70, 80, 90, None],
    ]


def test_table_rstrip_5():
    table = Table("empty")
    expected = table.serialize()
    table.rstrip()
    result = table.serialize()
    assert result == expected



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.shapes import DrawGroup


@pytest.fixture
def group() -> Iterable[DrawGroup]:
    shape = DrawGroup(
        name="some name",
        style="style name",
        draw_id="draw_id",
        svg_y="30cm",
        presentation_class="pres class",
        presentation_style="pres style",
        caption_id="capid",
        class_names="class1",
        z_index=4,
        end_cell_address="x10",
        end_x="10",
        end_y="20",
        table_background=True,
        anchor_type="page",
        anchor_page=42,
        xml_id="xmlid",
    )
    shape.svg_title = "title"
    yield shape


def test_drawgroup_minimal():
    group = DrawGroup()
    assert group._canonicalize() == "<draw:g></draw:g>"


def test_drawgroup_class():
    group = Element.from_tag("<draw:g/>")
    assert isinstance(group, DrawGroup)


def test_drawgroup_complete_1(group):
    assert group.name == "some name"


def test_drawgroup_complete_2(group):
    assert group.style == "style name"


def test_drawgroup_complete_3(group):
    assert group.draw_id == "draw_id"


def test_drawgroup_complete_4(group):
    assert group.svg_y == "30cm"


def test_drawgroup_complete_5(group):
    assert group.presentation_class == "pres class"


def test_drawgroup_complete_6(group):
    assert group.presentation_style == "pres style"


def test_drawgroup_complete_7(group):
    assert group.caption_id == "capid"


def test_drawgroup_complete_8(group):
    assert group.class_names == "class1"


def test_drawgroup_complete_9(group):
    assert group.z_index == 4


def test_drawgroup_complete_10(group):
    assert group.end_cell_address == "x10"


def test_drawgroup_complete_11(group):
    assert group.end_x == "10"


def test_drawgroup_complete_12(group):
    assert group.end_y == "20"


def test_drawgroup_complete_13(group):
    assert group.table_background is True


def test_drawgroup_complete_14(group):
    assert group.anchor_type == "page"


def test_drawgroup_complete_15(group):
    assert group.anchor_page == 42


def test_drawgroup_complete_16(group):
    assert group.xml_id == "xmlid"


def test_drawgroup_complete_formatted_text(group):
    assert group.get_formatted_text() == "title\n\n"
